@@ -1,4 +1,4 @@
-# Copyright 2004-2020 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -85,7 +85,7 @@ def float_or_none(x):
     return float(x)
 
 
-def matrixcolor(x):
+def matrix(x):
     if x is None:
         return None
     elif callable(x):
@@ -102,61 +102,8 @@ def mesh(x):
 
 
 # A dictionary giving property names and the corresponding default
-# values.
-PROPERTIES = {
-    "pos" : (position, position),
-    "xpos" : position,
-    "ypos" : position,
-    "anchor" : (position, position),
-    "xanchor" : position,
-    "yanchor" : position,
-    "xaround" : position,
-    "yaround" : position,
-    "xanchoraround" : float,
-    "yanchoraround" : float,
-    "align" : (float, float),
-    "xalign" : float,
-    "yalign" : float,
-    "rotate" : float,
-    "rotate_pad" : bool,
-    "transform_anchor" : bool,
-    "xzoom" : float,
-    "yzoom" : float,
-    "zoom" : float,
-    "nearest" : bool_or_none,
-    "alpha" : float,
-    "additive" : float,
-    "around" : (position, position),
-    "alignaround" : (float, float),
-    "angle" : float,
-    "radius" : float,
-    "crop" : (float, float, float, float),
-    "crop_relative" : bool,
-    "xsize" : int,
-    "ysize" : int,
-    "size" : (int, int),
-    "fit" : str,
-    "maxsize" : (int, int),
-    "corner1" : (float, float),
-    "corner2" : (float, float),
-    "subpixel" : bool,
-    "delay" : float,
-    "xoffset" : float,
-    "yoffset" : float,
-    "offset" : (int, int),
-    "xcenter" : position,
-    "ycenter" : position,
-    "debug" : any_object,
-    "events" : bool,
-    "xpan" : float_or_none,
-    "ypan" : float_or_none,
-    "xtile" : int,
-    "ytile" : int,
-    "matrixcolor" : matrixcolor,
-    "shader" : any_object,
-    "mesh" : mesh,
-    "blur" : float_or_none,
-    }
+# values. This is massively added to by renpy.display.transform.
+PROPERTIES = { }
 
 
 def correct_type(v, b, ty):
@@ -389,12 +336,19 @@ class ATLTransformBase(renpy.object.Object):
         self.parent_transform = None
 
         # The offset between st and when this ATL block first executed.
-        self.atl_st_offset = 0
+        if renpy.config.atl_start_on_show:
+            self.atl_st_offset = None
+        else:
+            self.atl_st_offset = 0
 
         if renpy.game.context().init_phase:
             compile_queue.append(self)
 
     def _handles_event(self, event):
+
+        if (event == "replaced") and (self.atl_state is None):
+            return True
+
         if (self.block is not None) and (self.block._handles_event(event)):
             return True
 
@@ -424,6 +378,7 @@ class ATLTransformBase(renpy.object.Object):
         super(ATLTransformBase, self).take_execution_state(t)
 
         self.atl_st_offset = None
+        self.atl_state = None
 
         if self is t:
             return
@@ -451,7 +406,6 @@ class ATLTransformBase(renpy.object.Object):
         self.at = t.at
         self.st_offset = t.st_offset
         self.at_offset = t.at_offset
-
         self.atl_st_offset = t.atl_st_offset
 
         if self.child is renpy.display.motion.null:
@@ -862,12 +816,16 @@ class Block(Statement):
                     loop_end = target - arg
                     duration = loop_end - loop_start
 
-                    if duration <= 0:
+                    if (state is None) and (duration <= 0):
                         raise Exception("ATL appears to be in an infinite loop.")
 
                     # Figure how many durations can occur between the
                     # start of the loop and now.
-                    new_repeats = int((target - loop_start) / duration)
+
+                    if duration:
+                        new_repeats = int((target - loop_start) / duration)
+                    else:
+                        new_repeats = 0
 
                     if count is not None:
                         if repeats + new_repeats >= count:
@@ -1184,12 +1142,27 @@ class Interpolation(Statement):
 
         warper = warpers.get(self.warper, self.warper)
 
-        if (self.warper != "instant") and (state is None) and (
-                (trans.atl_state is not None) or (trans.st == 0)
-                ):
-            first = True
+        if trans.atl.animation:
+            st_or_at = trans.at
         else:
-            first = False
+            st_or_at = trans.st
+
+        # True if we want want to make sure this interpolation is shown for at
+        # least one frame.
+        if self.warper == "instant":
+            first_frame = False
+        elif state is not None:
+            first_frame = False
+        elif (self.duration == 0) and (not self.properties and not self.revolution and not self.splines):
+            first_frame = True
+        elif trans.atl_state is not None:
+            first_frame = True
+        elif st_or_at == 0:
+            first_frame = True
+        else:
+            # This is the case when we're skipping through a displayable to
+            # find the right time.
+            first_frame = False
 
         if self.duration:
             complete = min(1.0, st / self.duration)
@@ -1313,7 +1286,7 @@ class Interpolation(Statement):
             value = interpolate_spline(complete, values)
             setattr(trans.state, name, value)
 
-        if ((not first) or (not renpy.config.atl_one_frame)) and (st >= self.duration):
+        if (st >= self.duration) and ((not first_frame) or (not renpy.config.atl_one_frame)):
             return "next", st - self.duration, None
         else:
             if not self.properties and not self.revolution and not self.splines:
@@ -1604,15 +1577,22 @@ class On(Statement):
         # handle it.
         for event in events:
 
-            if event in self.handlers:
+            while event:
+                if event in self.handlers:
+                    break
 
-                # Do not allow people to abort the hide or replaced event.
-                lock_event = (name == "hide" and trans.hide_request) or (name == "replaced" and trans.replaced_request)
+                event = event.partition("_")[2]
 
-                if not lock_event:
-                    name = event
-                    start = st
-                    cstate = None
+            if not event:
+                continue
+
+            # Do not allow people to abort the hide or replaced event.
+            lock_event = (name == "hide" and trans.hide_request) or (name == "replaced" and trans.replaced_request)
+
+            if not lock_event:
+                name = event
+                start = st
+                cstate = None
 
         while True:
 
