@@ -1,4 +1,4 @@
-# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,7 +19,9 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+
 import os.path
 import sys
 import subprocess
@@ -27,19 +29,21 @@ import io
 
 FSENCODING = sys.getfilesystemencoding() or "utf-8"
 
-# Sets the default encoding to the filesystem encoding.
+# Sets the default encoding to utf-8.
 old_stdout = sys.stdout
 old_stderr = sys.stderr
 
-reload(sys)
-sys.setdefaultencoding(FSENCODING)  # @UndefinedVariable
+if PY2:
+    sys_executable = sys.executable
+    reload(sys)
+    sys.setdefaultencoding("utf-8") # @UndefinedVariable
+    sys.executable = sys_executable
 
 sys.stdout = old_stdout
 sys.stderr = old_stderr
 
 import renpy.error
 
-
 # Extra things used for distribution.
 
 
@@ -66,7 +70,7 @@ def extra_imports():
     import difflib; difflib
     import shutil; shutil
     import tarfile; tarfile
-    import bz2; bz2  # @UnresolvedImport
+    import bz2; bz2 # @UnresolvedImport
     import webbrowser; webbrowser
     import posixpath; posixpath
     import ctypes; ctypes
@@ -87,6 +91,8 @@ def extra_imports():
     import pstats; pstats
     import _ssl; _ssl
     import SimpleHTTPServer; SimpleHTTPServer
+    import wave; wave
+    import sunau; sunau
 
     # Used by requests.
     import cgi; cgi
@@ -121,14 +127,13 @@ def null_files():
 
 null_files()
 
-
 trace_file = None
 trace_local = None
 
 
 def trace_function(frame, event, arg):
     fn = os.path.basename(frame.f_code.co_filename)
-    print(fn, frame.f_lineno, frame.f_code.co_name, event, file=trace_file)
+    trace_file.write("{} {} {} {}\n".format(fn, frame.f_lineno, frame.f_code.co_name, event))
     return trace_local
 
 
@@ -136,7 +141,7 @@ def enable_trace(level):
     global trace_file
     global trace_local
 
-    trace_file = file("trace.txt", "w", 1)
+    trace_file = open("trace.txt", "w", buffering=1, encoding="utf-8")
 
     if level > 1:
         trace_local = trace_function
@@ -158,21 +163,24 @@ def popen_del(self, *args, **kwargs):
 
 def bootstrap(renpy_base):
 
-    global renpy  # W0602
+    global renpy # W0602
 
-    import renpy.log  # @UnusedImport
+    import renpy.log # @UnusedImport
 
     # Remove a legacy environment setting.
-    if os.environ.get(b"SDL_VIDEODRIVER", "") == "windib":
-        del os.environ[b"SDL_VIDEODRIVER"]
+    if os.environ.get("SDL_VIDEODRIVER", "") == "windib":
+        del os.environ["SDL_VIDEODRIVER"]
 
-    renpy_base = unicode(renpy_base, FSENCODING, "replace")
+    if not isinstance(renpy_base, str):
+        renpy_base = str(renpy_base, FSENCODING)
 
     # If environment.txt exists, load it into the os.environ dictionary.
     if os.path.exists(renpy_base + "/environment.txt"):
         evars = { }
-        execfile(renpy_base + "/environment.txt", evars)
-        for k, v in evars.iteritems():
+        with open(renpy_base + "/environment.txt", "r") as f:
+            code = compile(f.read(), renpy_base + "/environment.txt", 'exec')
+            exec(code, evars)
+        for k, v in evars.items():
             if k not in os.environ:
                 os.environ[k] = str(v)
 
@@ -180,12 +188,14 @@ def bootstrap(renpy_base):
     # .app file.), if on a mac.
     alt_path = os.path.abspath("renpy_base")
     if ".app" in alt_path:
-        alt_path = alt_path[:alt_path.find(".app")+4]
+        alt_path = alt_path[:alt_path.find(".app") + 4]
 
         if os.path.exists(alt_path + "/environment.txt"):
             evars = { }
-            execfile(alt_path + "/environment.txt", evars)
-            for k, v in evars.iteritems():
+            with open(alt_path + "/environment.txt", "rb") as f:
+                code = compile(f.read(), alt_path + "/environment.txt", 'exec')
+                exec(code, evars)
+            for k, v in evars.items():
                 if k not in os.environ:
                     os.environ[k] = str(v)
 
@@ -203,7 +213,9 @@ def bootstrap(renpy_base):
         enable_trace(args.trace)
 
     if args.basedir:
-        basedir = os.path.abspath(args.basedir).decode(FSENCODING)
+        basedir = os.path.abspath(args.basedir)
+        if not isinstance(basedir, str):
+            basedir = basedir.decode(FSENCODING)
     else:
         basedir = renpy_base
 
@@ -211,6 +223,11 @@ def bootstrap(renpy_base):
         sys.stderr.write("Base directory %r does not exist. Giving up.\n" % (basedir,))
         sys.exit(1)
 
+    # Make game/ on Android.
+    if renpy.android:
+        if not os.path.exists(basedir + "/game"):
+            os.mkdir(basedir + "/game", 0o777)
+
     gamedirs = [ name ]
     game_name = name
 
@@ -267,7 +284,7 @@ You may be using a system install of python. Please run {0}.sh,
 
     # If we're not given a command, show the presplash.
     if args.command == "run" and not renpy.mobile:
-        import renpy.display.presplash  # @Reimport
+        import renpy.display.presplash # @Reimport
         renpy.display.presplash.start(basedir, gamedir)
 
     # Ditto for the Ren'Py module.
@@ -285,7 +302,7 @@ You may be using a system install of python. Please run {0}.sh,
 
     # Load up all of Ren'Py, in the right order.
 
-    import renpy  # @Reimport
+    import renpy # @Reimport
     renpy.import_all()
 
     renpy.loader.init_importer()
@@ -348,6 +365,8 @@ You may be using a system install of python. Please run {0}.sh,
         if "RENPY_SHUTDOWN_TRACE" in os.environ:
             enable_trace(int(os.environ["RENPY_SHUTDOWN_TRACE"]))
 
+        renpy.display.tts.tts(None)
+
         renpy.display.im.cache.quit()
 
         if renpy.display.draw:
