@@ -1,4 +1,4 @@
-# Copyright 2004-2020 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -76,16 +76,17 @@ def compile_event(key, keydown):
         else:
             return "(False)"
 
-    # Deal with the Joystick / Gamepad.
-    if part[0] == "joy" or part[0] == "pad":
-        return "(False)"
 
-    MODIFIERS = { "keydown", "keyup", "repeat", "alt", "meta", "shift", "noshift", "ctrl" }
+    MODIFIERS = { "keydown", "keyup", "repeat", "alt", "meta", "shift", "noshift", "ctrl", "osctrl" }
     modifiers = set()
 
     while part[0] in MODIFIERS:
         modifiers.add(part.pop(0))
 
+    # Deal with the Joystick / Gamepad.
+    if part[0] == "joy" or part[0] == "pad":
+        return "(False)"
+
     key = "_".join(part)
 
     if "keydown" in modifiers:
@@ -106,7 +107,7 @@ def compile_event(key, keydown):
 
     if key not in [ "K_LALT", "K_RALT" ]:
 
-        if "alt" in modifiers:
+        if "alt" in modifiers or (renpy.macintosh and "osctrl" in modifiers):
             rv += " and (ev.mod & %d)" % pygame.KMOD_ALT
         else:
             rv += " and not (ev.mod & %d)" % pygame.KMOD_ALT
@@ -120,7 +121,7 @@ def compile_event(key, keydown):
 
     if key not in [ "K_LCTRL", "K_RCTRL" ]:
 
-        if "ctrl" in modifiers:
+        if "ctrl" in modifiers or (not renpy.macintosh and "osctrl" in modifiers):
             rv += " and (ev.mod & %d)" % pygame.KMOD_CTRL
         else:
             rv += " and not (ev.mod & %d)" % pygame.KMOD_CTRL
@@ -224,6 +225,9 @@ def map_event(ev, keysym):
 
         return False
 
+    if isinstance(keysym, list):
+        keysym = tuple(keysym)
+
     check_code = event_cache.get(keysym, None)
     if check_code is None:
         check_code = eval("lambda ev : " + compile_event(keysym, True), globals())
@@ -232,17 +236,20 @@ def map_event(ev, keysym):
     return check_code(ev)
 
 
-def map_keyup(ev, name):
+def map_keyup(ev, keysym):
     """Returns true if the event matches the named keycode being released."""
 
     if ev.type == renpy.display.core.EVENTNAME:
-        if (name in ev.eventnames) and ev.up:
+        if (keysym in ev.eventnames) and ev.up:
             return True
 
-    check_code = keyup_cache.get(name, None)
+    if isinstance(keysym, list):
+        keysym = tuple(keysym)
+
+    check_code = keyup_cache.get(keysym, None)
     if check_code is None:
-        check_code = eval("lambda ev : " + compile_event(name, False), globals())
-        keyup_cache[name] = check_code
+        check_code = eval("lambda ev : " + compile_event(keysym, False), globals())
+        keyup_cache[keysym] = check_code
 
     return check_code(ev)
 
@@ -506,6 +513,9 @@ class PauseBehavior(renpy.display.layout.Null):
 
     def event(self, ev, x, y, st):
 
+        if ev.type == renpy.display.core.TIMEEVENT and ev.modal:
+            return
+
         if st >= self.delay:
 
             if self.voice and renpy.config.nw_voice:
@@ -627,7 +637,7 @@ class SayBehavior(renpy.display.layout.Null):
 
                 if not renpy.config.enable_rollback_side:
                     rollback_side = "disable"
-                if renpy.exports.mobile:
+                if renpy.exports.variant("mobile"):
                     rollback_side = renpy.game.preferences.mobile_rollback_side
                 else:
                     rollback_side = renpy.game.preferences.desktop_rollback_side
@@ -667,7 +677,11 @@ class SayBehavior(renpy.display.layout.Null):
         if renpy.config.skipping and renpy.config.allow_skipping and renpy.store._skipping:
 
             if ev.type == renpy.display.core.TIMEEVENT and st >= skip_delay:
-                if renpy.game.preferences.skip_unseen:
+
+                if ev.modal:
+                    renpy.config.skipping = None
+                    renpy.exports.restart_interaction()
+                elif renpy.game.preferences.skip_unseen:
                     return True
                 elif renpy.config.skipping == "fast":
                     return True
@@ -1120,6 +1134,49 @@ def input_post_per_interact():
             i.caret_pos = len(content)
 
 
+class CaretBlink(renpy.display.core.Displayable):
+    """
+    A displayable that renders the caret.
+    """
+
+    def __init__(self, caret, caret_blink, **properties):
+
+        properties.setdefault("yalign", 0.0)
+
+        super(CaretBlink, self).__init__(**properties)
+        caret = renpy.easy.displayable(caret)
+
+        if caret._duplicatable:
+            caret = caret._duplicate()
+            caret._unique()
+
+        self.caret = caret
+        self.caret_blink = caret_blink
+
+        self.st_base = 0
+
+    def get_position(self):
+        return self.caret.get_position()
+
+    def visit(self):
+        return [ self.caret ]
+
+    def render(self, width, height, st, at):
+        st -= self.st_base
+
+        cr = renpy.display.render.render(self.caret, width, height, st, at)
+        rv = renpy.display.render.Render(1, height)
+
+        ttl = self.caret_blink - st % self.caret_blink
+
+        if ttl > self.caret_blink / 2.0:
+            rv.blit(cr, (0, 0))
+
+        renpy.display.render.redraw(self, ttl % (self.caret_blink / 2.))
+
+        return rv
+
+
 class Input(renpy.text.text.Text): # @UndefinedVariable
     """
     This is a Displayable that takes text as input.
@@ -1136,6 +1193,8 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
     value = None
     shown = False
 
+    st = 0
+
     def __init__(self,
                  default="",
                  length=None,
@@ -1151,10 +1210,14 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
                  pixel_width=None,
                  value=None,
                  copypaste=False,
+                 caret_blink=None,
                  **properties):
 
         super(Input, self).__init__("", style=style, replaces=replaces, substitute=False, **properties)
 
+        if caret_blink is None:
+            caret_blink = renpy.config.input_caret_blink
+
         if value:
             self.value = value
             changed = value.set_text
@@ -1182,7 +1245,12 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
             if i.endswith("color"):
                 caretprops[i] = properties[i]
 
-        self.caret = renpy.display.image.Solid(xmaximum=1, style=style, **caretprops)
+        caret = renpy.display.image.Solid(xsize=1, style=style, **caretprops)
+
+        if caret_blink:
+            caret = CaretBlink(caret, caret_blink)
+
+        self.caret = caret
         self.caret_pos = len(self.content)
         self.old_caret_pos = self.caret_pos
 
@@ -1206,12 +1274,14 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
         old_content = self.content
 
         if new_content != self.content or editable != self.editable or edit:
+
             renpy.display.render.redraw(self, 0)
 
         self.editable = editable
 
         # Choose the caret.
         caret = self.style.caret
+
         if caret is None:
             caret = self.caret
 
@@ -1242,7 +1312,7 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
         def set_content(content):
 
             if content == "":
-                content = u"\u200b"
+                content = u" "
 
             if editable:
                 l = len(content)
@@ -1251,6 +1321,10 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
             else:
                 self.set_text([self.prefix, content.replace("{", "{{"), self.suffix ])
 
+            if isinstance(self.caret, CaretBlink):
+                self.caret.st_base = self.st
+                renpy.display.render.redraw(self.caret, 0)
+
         set_content(new_content)
 
         if check_size and self.pixel_width:
@@ -1305,12 +1379,20 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
 
     def event(self, ev, x, y, st):
 
+        self.st = st
         self.old_caret_pos = self.caret_pos
 
         if not self.editable:
             return None
 
-        if (ev.type == pygame.KEYDOWN) and (pygame.key.get_mods() & pygame.KMOD_LALT) and (not ev.unicode):
+        edit_controls = any([
+            map_event(ev, "input_jump_word_left"),
+            map_event(ev, "input_jump_word_right"),
+            map_event(ev, "input_delete_word"),
+            map_event(ev, "input_delete_full"),
+        ])
+
+        if (ev.type == pygame.KEYDOWN) and (pygame.key.get_mods() & pygame.KMOD_LALT) and (not ev.unicode) and not edit_controls:
             return None
 
         l = len(self.content)
@@ -1348,6 +1430,21 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
             renpy.display.render.redraw(self, 0)
             raise renpy.display.core.IgnoreEvent()
 
+        elif map_event(ev, "input_jump_word_left"):
+            if self.caret_pos > 0:
+                space_pos = 0
+                prev_end = 0
+                for item in re.finditer(r"\s+", self.content[:self.caret_pos]):
+                    start, end = item.span()
+                    if end != self.caret_pos:
+                        space_pos = end
+                    prev_end = end
+                self.caret_pos = space_pos
+                self.update_text(self.content, self.editable)
+
+            renpy.display.render.redraw(self, 0)
+            raise renpy.display.core.IgnoreEvent()
+
         elif map_event(ev, "input_right"):
             if self.caret_pos < l:
                 self.caret_pos += 1
@@ -1356,6 +1453,19 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
             renpy.display.render.redraw(self, 0)
             raise renpy.display.core.IgnoreEvent()
 
+        elif map_event(ev, "input_jump_word_right"):
+            if self.caret_pos < l:
+                space_pos = l
+                for item in re.finditer(r"\s+", self.content[self.caret_pos + 1:]):
+                    start, end = item.span()
+                    space_pos = end
+                    break
+                self.caret_pos = min(space_pos + self.caret_pos + 1, l)
+                self.update_text(self.content, self.editable)
+
+            renpy.display.render.redraw(self, 0)
+            raise renpy.display.core.IgnoreEvent()
+
         elif map_event(ev, "input_delete"):
             if self.caret_pos < l:
                 content = self.content[0:self.caret_pos] + self.content[self.caret_pos + 1:l]
@@ -1364,6 +1474,29 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
             renpy.display.render.redraw(self, 0)
             raise renpy.display.core.IgnoreEvent()
 
+        elif map_event(ev, "input_delete_word"):
+            if self.caret_pos <= l:
+                space_pos = 0
+                for item in re.finditer(r"\s+", self.content[:self.caret_pos]):
+                    start, end = item.span()
+                    if end != self.caret_pos:
+                        space_pos = end
+                content = self.content[0:space_pos] + self.content[self.caret_pos:l]
+                self.caret_pos = space_pos
+                self.update_text(content, self.editable)
+
+            renpy.display.render.redraw(self, 0)
+            raise renpy.display.core.IgnoreEvent()
+
+        elif map_event(ev, "input_delete_full"):
+            if self.caret_pos <= l:
+                content = self.content[self.caret_pos:l]
+                self.caret_pos = 0
+                self.update_text(content, self.editable)
+
+            renpy.display.render.redraw(self, 0)
+            raise renpy.display.core.IgnoreEvent()
+
         elif map_event(ev, "input_home"):
             self.caret_pos = 0
             self.update_text(self.content, self.editable)
@@ -1418,7 +1551,7 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
                 if self.allow:
 
                     # Allow is regex
-                    if isinstance(self.allow, re._pattern_type):
+                    if isinstance(self.allow, re.Pattern):
 
                         # Character doesn't match
                         if self.allow.search(c) is None:
@@ -1432,7 +1565,7 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
                 if self.exclude:
 
                     # Exclude is regex
-                    if isinstance(self.exclude, re._pattern_type):
+                    if isinstance(self.exclude, re.Pattern):
 
                         # Character matches
                         if self.exclude.search(c) is not None:
@@ -1458,6 +1591,8 @@ class Input(renpy.text.text.Text): # @UndefinedVariable
             raise renpy.display.core.IgnoreEvent()
 
     def render(self, width, height, st, at):
+        self.st = st
+
         rv = super(Input, self).render(width, height, st, at)
 
         if self.editable:
@@ -1532,6 +1667,9 @@ class Adjustment(renpy.object.Object):
             This function is called with the adjustment object when
             the range of the adjustment is set by a viewport.
 
+            This function may be called multiple times, as part of the layout
+            process.
+
         `force_step`
             If True and this adjustment changes by dragging associated
             viewport or a bar, value will be changed only if the drag
@@ -1694,6 +1832,7 @@ class Bar(renpy.display.core.Displayable):
                  replaces=None,
                  hovered=None,
                  unhovered=None,
+                 released=None,
                  **properties):
 
         self.value = None
@@ -1747,6 +1886,8 @@ class Bar(renpy.display.core.Displayable):
         self.hovered = hovered
         self.unhovered = unhovered
 
+        self.released = released
+
     def per_interact(self):
         if self.value is not None:
             adjustment = self.value.get_adjustment()
@@ -1773,11 +1914,12 @@ class Bar(renpy.display.core.Displayable):
             if redraw is not None:
                 renpy.display.render.redraw(self, redraw)
 
-        xminimum = self.style.xminimum
-        yminimum = self.style.yminimum
+        xminimum, yminimum = renpy.display.layout.xyminimums(self.style, width, height)
 
         if xminimum is not None:
             width = max(width, xminimum)
+
+        if yminimum is not None:
             height = max(height, yminimum)
 
         # Store the width and height for the event function to use.
@@ -1935,6 +2077,7 @@ class Bar(renpy.display.core.Displayable):
             just_grabbed = True
             grabbed = True
             ignore_event = True
+            renpy.exports.play(self.style.activate_sound)
 
         if grabbed:
 
@@ -2004,6 +2147,10 @@ class Bar(renpy.display.core.Displayable):
                 if rv is not None:
                     return rv
 
+            rv = run(self.released)
+            if rv is not None:
+                return rv
+
             raise renpy.display.core.IgnoreEvent()
 
         if value != old_value:
@@ -2070,13 +2217,11 @@ class Conditional(renpy.display.layout.Container):
             return self.child.event(ev, x, y, st)
 
 
-class TimerState(renpy.python.RevertableObject):
+class TimerState(renpy.python.AlwaysRollback):
     """
     Stores the state of the timer, which may need to be rolled back.
     """
 
-    # Prevents us from having to worry about our initialization being
-    # rolled back.
     started = False
     next_event = None
 
@@ -2116,13 +2261,16 @@ class Timer(renpy.display.layout.Null):
         # Did we start the timer?
         self.started = False
 
-        if replaces is not None:
+        if isinstance(replaces, Timer):
             self.state = replaces.state
         else:
             self.state = TimerState()
 
     def event(self, ev, x, y, st):
 
+        if ev.type == renpy.display.core.TIMEEVENT and ev.modal:
+            return
+
         state = self.state
 
         if not state.started:
@@ -2162,7 +2310,7 @@ class MouseArea(renpy.display.core.Displayable):
         # Are we hovered right now?
         self.is_hovered = False
 
-        if replaces is not None:
+        if isinstance(replaces, MouseArea):
             self.is_hovered = replaces.is_hovered
 
         # Taken from the render.
@@ -2183,14 +2331,21 @@ class MouseArea(renpy.display.core.Displayable):
         if renpy.display.focus.get_grab():
             return
 
-        if self.style.focus_mask is not None:
-            crend = renpy.display.render.render(self.style.focus_mask, self.width, self.height, st, self.at_st_offset + st)
-            is_hovered = crend.is_pixel_opaque(x, y)
-        elif 0 <= x < self.width and 0 <= y < self.height:
-            is_hovered = True
-        else:
+        if renpy.display.focus.pending_focus_type == 'keyboard':
             is_hovered = False
 
+        if (ev.type == renpy.display.core.TIMEEVENT) and ev.modal:
+            is_hovered = False
+
+        else:
+            if self.style.focus_mask is not None:
+                crend = renpy.display.render.render(self.style.focus_mask, self.width, self.height, st, self.at_st_offset + st)
+                is_hovered = crend.is_pixel_opaque(x, y)
+            elif 0 <= x < self.width and 0 <= y < self.height:
+                is_hovered = True
+            else:
+                is_hovered = False
+
         if is_hovered and not self.is_hovered:
             self.is_hovered = True
 
