@@ -1,4 +1,4 @@
-# Copyright 2004-2020 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -410,7 +410,7 @@ class ImageReference(renpy.display.core.Displayable):
 
         except Exception as e:
 
-            if renpy.config.debug:
+            if renpy.config.raise_image_exceptions and (renpy.config.debug or renpy.config.developer):
                 raise
 
             error(str(e))
@@ -657,12 +657,13 @@ class DynamicImage(renpy.display.core.Displayable):
         if not update:
             return True
 
-        self.raw_target = target
+        raw_target = target
         old_target = self.target
 
         if target._duplicatable:
             target = target._duplicate(self._args)
 
+        self.raw_target = raw_target
         self.target = target
 
         renpy.display.render.redraw(self, 0)
@@ -904,6 +905,10 @@ class ShownImageInfo(renpy.object.Object):
         with that name couldn't be found.
         """
 
+        f = renpy.config.adjust_attributes.get(name[0], None) or renpy.config.adjust_attributes.get(None, None)
+        if f is not None:
+            name = f(name)
+
         if layer is None:
             layer = renpy.config.tag_layer.get(tag, "master")
 
