@@ -1,4 +1,4 @@
-# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -24,7 +24,8 @@
 # are None, at least to the point of making it through __init__. This is
 # so that prediction of images works.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
 
 import renpy.display
 from renpy.display.render import render
@@ -44,12 +45,12 @@ class Transition(renpy.display.core.Displayable):
     def event(self, ev, x, y, st):
 
         if self.events or ev.type == renpy.display.core.TIMEEVENT:
-            return self.new_widget.event(ev, x, y, st)  # E1101
+            return self.new_widget.event(ev, x, y, st) # E1101
         else:
             return None
 
     def visit(self):
-        return [ self.new_widget, self.old_widget ]  # E1101
+        return [ self.new_widget, self.old_widget ] # E1101
 
 
 def null_render(d, width, height, st, at):
@@ -300,6 +301,12 @@ class Pixellate(Transition):
         rv.operation = renpy.display.render.PIXELLATE
         rv.operation_parameter = 2 ** step
 
+        rv.mesh = True
+        rv.add_shader("renpy.texture")
+        rv.add_property("texture_scaling", "nearest_mipmap_nearest")
+        rv.add_property("anisotropic", False)
+        rv.add_uniform("u_lod_bias", step + 1)
+
         renpy.display.render.redraw(self, 0)
 
         return rv
@@ -369,6 +376,20 @@ class Dissolve(Transition):
         rv.operation_alpha = self.alpha or renpy.config.dissolve_force_alpha
         rv.operation_complete = complete
 
+        if renpy.display.render.models:
+
+            target = rv.get_size()
+
+            if top.get_size() != target:
+                top = top.subsurface((0, 0, width, height))
+            if bottom.get_size() != target:
+                bottom = bottom.subsurface((0, 0, width, height))
+
+            rv.mesh = True
+            rv.add_shader("renpy.dissolve")
+            rv.add_uniform("u_renpy_dissolve", complete)
+            rv.add_property("mipmap", renpy.config.mipmap_dissolves if (self.style.mipmap is None) else self.style.mipmap)
+
         rv.blit(bottom, (0, 0), focus=False, main=False)
         rv.blit(top, (0, 0), focus=True, main=True)
 
@@ -388,9 +409,10 @@ class ImageDissolve(Transition):
     dissolve in first, and black pixels will dissolve in last.
 
     `image`
-        A control image to use. This must be either an image file or
-        image manipulator. The control image should be the size of
-        the scenes being dissolved.
+        A control image to use. If `config.gl2` is True, then this can be any displayable,
+        else it needs to be either an image file or an image manipulator. The control image
+        should be the size of the scenes being dissolved, and if `reverse=True`,
+        it should be fully opaque.
 
     `time`
         The time the dissolve will take.
@@ -452,25 +474,39 @@ class ImageDissolve(Transition):
         self.alpha = alpha
         self.time_warp = time_warp
 
-        if not reverse:
-
-            # Copies red -> alpha
-            matrix = renpy.display.im.matrix(
-                0, 0, 0, 0, 1,
-                0, 0, 0, 0, 1,
-                0, 0, 0, 0, 1,
-                1, 0, 0, 0, 0)
+        if renpy.display.render.models:
+            if not reverse:
+                # Copies red -> alpha
+                matrix = renpy.display.matrix.Matrix([0, 0, 0, 0,
+                                                      0, 0, 0, 0,
+                                                      0, 0, 0, 0,
+                                                      1, 0, 0, 0, ])
+            else:
+                # Copies alpha-red -> alpha
+                matrix = renpy.display.matrix.Matrix([0, 0, 0, 0,
+                                                      0, 0, 0, 0,
+                                                      0, 0, 0, 0,
+                                                      -1, 0, 0, 1, ])
+                # asserts the displayable is not transparent at all
+                # just like InvertMatrix does
+            self.image = renpy.display.motion.Transform(image, matrixcolor=matrix)
 
         else:
-
-            # Copies 1-red -> alpha
-            matrix = renpy.display.im.matrix(
-                0, 0, 0, 0, 1,
-                0, 0, 0, 0, 1,
-                0, 0, 0, 0, 1,
-                - 1, 0, 0, 0, 1)
-
-        self.image = renpy.display.im.MatrixColor(image, matrix)
+            if not reverse:
+                # Copies red -> alpha
+                matrix = renpy.display.im.matrix(
+                    0, 0, 0, 0, 1,
+                    0, 0, 0, 0, 1,
+                    0, 0, 0, 0, 1,
+                    1, 0, 0, 0, 0)
+            else:
+                # Copies 1-red -> alpha
+                matrix = renpy.display.im.matrix(
+                    0, 0, 0, 0, 1,
+                    0, 0, 0, 0, 1,
+                    0, 0, 0, 0, 1,
+                    -1, 0, 0, 0, 1)
+            self.image = renpy.display.im.MatrixColor(image, matrix)
 
         if ramp is not None:
             ramplen = len(ramp)
@@ -509,6 +545,35 @@ class ImageDissolve(Transition):
         rv.operation_complete = complete
         rv.operation_parameter = self.ramplen
 
+        if renpy.display.render.models:
+
+            target = rv.get_size()
+
+            if image.get_size() != target:
+                image = image.subsurface((0, 0, width, height))
+            if top.get_size() != target:
+                top = top.subsurface((0, 0, width, height))
+            if bottom.get_size() != target:
+                bottom = bottom.subsurface((0, 0, width, height))
+
+            ramp = self.ramplen
+
+            # Prevent a DBZ if the user gives us a 0 ramp.
+            if ramp < 1:
+                ramp = 1
+
+            # Compute the offset to apply to the alpha.
+            start = -1.0
+            end = ramp / 256.0
+            offset = start + (end - start) * complete
+
+            rv.mesh = True
+
+            rv.add_shader("renpy.imagedissolve",)
+            rv.add_uniform("u_renpy_dissolve_offset", offset)
+            rv.add_uniform("u_renpy_dissolve_multiplier", 256.0 / ramp)
+            rv.add_property("mipmap", renpy.config.mipmap_dissolves if (self.style.mipmap is None) else self.style.mipmap)
+
         rv.blit(image, (0, 0), focus=False, main=False)
         rv.blit(bottom, (0, 0), focus=False, main=False)
         rv.blit(top, (0, 0), focus=True, main=True)
@@ -543,6 +608,8 @@ class AlphaDissolve(Transition):
         new image.
      """
 
+    mipmap = None
+
     def __init__(
             self,
             control,
@@ -591,6 +658,13 @@ class AlphaDissolve(Transition):
         rv.operation_complete = 256.0 / (256.0 + 256.0)
         rv.operation_parameter = 256
 
+        if renpy.display.render.models:
+            rv.mesh = True
+            rv.add_shader("renpy.imagedissolve",)
+            rv.add_uniform("u_renpy_dissolve_offset", 0)
+            rv.add_uniform("u_renpy_dissolve_multiplier", 1.0)
+            rv.add_property("mipmap", renpy.config.mipmap_dissolves if (self.style.mipmap is None) else self.style.mipmap)
+
         rv.blit(control, (0, 0), focus=False, main=False)
 
         if not self.reverse:
