@@ -1,4 +1,4 @@
-# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,7 +19,9 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+
 import renpy.display
 import renpy.text
 import codecs
@@ -28,10 +30,9 @@ import re
 import sys
 import collections
 import textwrap
+import builtins
 
-import __builtin__
-
-python_builtins = set(dir(__builtin__))
+python_builtins = set(dir(builtins))
 renpy_builtins = set()
 
 image_prefixes = None
@@ -51,6 +52,13 @@ image_prefixes = None
 # The node the report will be about:
 report_node = None
 
+# Collect define/default statements to check for duplication
+all_define_statments = {}
+all_default_statements = {}
+
+# True if at east one error was reported, false otherwise.
+error_reported = False
+
 # Reports a message to the user.
 
 
@@ -61,8 +69,11 @@ def report(msg, *args):
         out = ""
 
     out += msg % args
-    print()
-    print(out.encode('utf-8'))
+    print("")
+    print(out)
+
+    global error_reported
+    error_reported = True
 
 
 added = { }
@@ -71,10 +82,11 @@ added = { }
 # occurs.
 
 
-def add(msg):
+def add(msg, *args):
     if not msg in added:
         added[msg] = True
-        print(unicode(msg).encode('utf-8'))
+        msg = str(msg) % args
+        print(msg)
 
 
 # Tries to evaluate an expression, announcing an error if it fails.
@@ -238,8 +250,8 @@ def image_exists_precise(name):
 
             try:
                 da = renpy.display.core.DisplayableArguments()
-                da.name=( im[0], ) + tuple(i for i in name[1:] if i in attrs)
-                da.args=tuple(i for i in name[1:] if i in rest)
+                da.name = (im[0],) + tuple(i for i in name[1:] if i in attrs)
+                da.args = tuple(i for i in name[1:] if i in rest)
                 da.lint = True
                 d._duplicate(da)
             except:
@@ -259,6 +271,11 @@ def image_exists(name, expression, tag, precise=True):
     Checks a scene or show statement for image existence.
     """
 
+    orig = name
+    f = renpy.config.adjust_attributes.get(name[0], None) or renpy.config.adjust_attributes.get(None, None)
+    if f is not None:
+        name = f(name)
+
     # Add the tag to the set of known tags.
     tag = tag or name[0]
     image_prefixes[tag] = True
@@ -275,7 +292,7 @@ def image_exists(name, expression, tag, precise=True):
     if image_exists_precise(name):
         return
 
-    report("'{}' is not an image.".format(" ".join(name)) )
+    report("'%s' is not an image.", " ".join(orig))
 
 
 # Only check each file once.
@@ -363,9 +380,9 @@ def precheck_show(node):
     tag = imspec(node.imspec)[2]
     image_prefixes[tag] = True
 
-
 # Lints ast.Hide.
 
+
 def check_hide(node):
 
     name, _expression, tag, _at_list, layer, _zorder, _behind = imspec(node.imspec)
@@ -402,10 +419,21 @@ def check_user(node):
         report("Didn't properly report what the next statement should be.")
 
 
+def quote_text(s):
+
+    s = s.replace("\\", "\\\\")
+    s = s.replace("\"", "\\\"")
+    s = s.replace("\t", "\\t")
+    s = s.replace("\n", "\\n")
+
+    return "\"" + s + "\""
+
+
 def text_checks(s):
+
     msg = renpy.text.extras.check_text_tags(s)
     if msg:
-        report("%s (in %s)", msg, repr(s)[1:])
+        report("%s (in %s)", msg, quote_text(s))
 
     if "%" in s and renpy.config.old_substitutions:
 
@@ -432,7 +460,7 @@ def text_checks(s):
                 elif c in "diouxXeEfFgGcrs%":
                     state = 0
                 else:
-                    report("Unknown string format code '%s' (in %s)", fmt, repr(s)[1:])
+                    report("Unknown string format code '%s' (in %s)", fmt, quote_text(s))
                     state = 0
 
             # In a mapping key.
@@ -442,7 +470,7 @@ def text_checks(s):
                     state = 1
 
         if state != 0:
-            report("Unterminated string format code '%s' (in %s)", fmt, repr(s)[1:])
+            report("Unterminated string format code '%s' (in %s)", fmt, quote_text(s))
 
 
 def check_say(node):
@@ -470,21 +498,22 @@ def check_say(node):
     if not isinstance(char, renpy.character.ADVCharacter):
         return
 
-    if node.attributes is None:
-        return
-
     if char.image_tag is None:
         return
 
-    name = (char.image_tag,) + node.attributes
+    for attributes in (node.attributes, node.temporary_attributes):
+        if attributes is None:
+            continue
 
-    if image_exists_imprecise(name):
-        return
+        name = (char.image_tag,) + attributes
 
-    if image_exists_imprecise(('side', ) + name):
-        return
+        if image_exists_imprecise(name):
+            continue
 
-    report("Could not find image (%s) corresponding to attributes on say statement.", " ".join(name))
+        if image_exists_imprecise(('side',) + name):
+            continue
+
+        report("Could not find image (%s) corresponding to attributes on say statement.", " ".join(name))
 
 
 def check_menu(node):
@@ -525,7 +554,6 @@ def check_while(node):
 
 
 def check_if(node):
-
     for condition, _block in node.entries:
         try_compile("in a condition of the if statement", condition)
 
@@ -544,6 +572,45 @@ def check_define(node, kind):
         report("'%s %s' replaces a Ren'Py built-in name, which may cause problems.", kind, node.varname)
 
 
+def check_redefined(node, kind):
+    """
+    Check if a define or default statement has already been created.
+    """
+    if kind == 'default':
+        scanned = all_default_statements
+    elif kind == 'define':
+        scanned = all_define_statments
+
+        if not (node.operator == "=" and node.index is None):
+            return
+
+    # Combine store name and varname
+
+    store_name = node.store
+    if store_name.startswith("store."):
+        store_name = store_name[6:]
+
+    if store_name:
+        full_name = "{}.{}".format(store_name, node.varname)
+    else:
+        full_name = node.varname
+
+    if full_name in renpy.config.lint_ignore_redefine:
+        return
+
+    original_node = scanned.get(full_name)
+    if original_node:
+        report(
+            "{} {} already defined at {}:{}".format(
+                kind,
+                full_name,
+                original_node.filename,
+                original_node.linenumber,
+            )
+        )
+    scanned[full_name] = node
+
+
 def check_style_property_displayable(name, property, d):
 
     if not d._duplicatable:
@@ -581,19 +648,18 @@ def check_style_property_displayable(name, property, d):
 def check_style(name, s):
 
     for p in s.properties:
-        for k, v in p.iteritems():
+        for k, v in p.items():
 
             # Treat font specially.
             if k.endswith("font"):
                 if isinstance(v, renpy.text.font.FontGroup):
-                    for f in v.fonts:
+                    for f in set(v.map.values()):
                         check_file(name, f)
                 else:
                     check_file(name, v)
 
             if isinstance(v, renpy.display.core.Displayable):
                 check_style_property_displayable(name, k, v)
-#                check_displayable(kname, v)
 
 
 def check_label(node):
@@ -618,12 +684,12 @@ def check_label(node):
 def check_screen(node):
 
     if (node.screen.parameters is None) and renpy.config.lint_screens_without_parameters:
-        report("The screen {} has not been given a parameter list.".format(node.screen.name))
-        add("This can be fixed by writing 'screen {}():' instead.".format(node.screen.name))
+        report("The screen %s has not been given a parameter list.", node.screen.name)
+        add("This can be fixed by writing 'screen %s():' instead.", node.screen.name)
 
 
 def check_styles():
-    for full_name, s in renpy.style.styles.iteritems():  # @UndefinedVariable
+    for full_name, s in renpy.style.styles.items(): # @UndefinedVariable
         name = "style." + full_name[0]
         for i in full_name[1:]:
             name += "[{!r}]".format(i)
@@ -703,17 +769,17 @@ def lint():
 
     ap = renpy.arguments.ArgumentParser(description="Checks the script for errors and prints script statistics.", require_command=False)
     ap.add_argument("filename", nargs='?', action="store", help="The file to write to.")
+    ap.add_argument("--error-code", action="store_true", help="If given, the error code is 0 if the game has no lint errros, 1 if lint errors are found.")
 
     args = ap.parse_args()
 
     if args.filename:
-        f = open(args.filename, "w")
+        f = open(args.filename, "w", encoding="utf-8")
         sys.stdout = f
 
     renpy.game.lint = True
 
-    print(codecs.BOM_UTF8)
-    print(unicode(renpy.version + " lint report, generated at: " + time.ctime()).encode("utf-8"))
+    print("\ufeff" + renpy.version + " lint report, generated at: " + time.ctime())
 
     # This supports check_hide.
     global image_prefixes
@@ -725,8 +791,8 @@ def lint():
     # Iterate through every statement in the program, processing
     # them. We sort them in filename, linenumber order.
 
-    all_stmts = [ (i.filename, i.linenumber, i) for i in renpy.game.script.all_stmts ]
-    all_stmts.sort()
+    all_stmts = list(renpy.game.script.all_stmts)
+    all_stmts.sort(key=lambda n : n.filename)
 
     # The current count.
     counts = collections.defaultdict(Count)
@@ -740,11 +806,11 @@ def lint():
 
     global report_node
 
-    for _fn, _ln, node in all_stmts:
+    for node in all_stmts:
         if isinstance(node, (renpy.ast.Show, renpy.ast.Scene)):
             precheck_show(node)
 
-    for _fn, _ln, node in all_stmts:
+    for node in all_stmts:
 
         if common(node):
             continue
@@ -806,9 +872,11 @@ def lint():
 
         elif isinstance(node, renpy.ast.Define):
             check_define(node, "define")
+            check_redefined(node, "define")
 
         elif isinstance(node, renpy.ast.Default):
             check_define(node, "default")
+            check_redefined(node, "default")
 
     report_node = None
 
@@ -843,10 +911,10 @@ characters per block. """.format(
 
         lines.append(s)
 
-    print()
-    print()
+    print("")
+    print("")
     print("Statistics:")
-    print()
+    print("")
 
     languages = list(counts)
     languages.sort()
@@ -858,19 +926,22 @@ characters per block. """.format(
 
     for l in lines:
         for ll in textwrap.wrap(l, 78):
-            print(ll.encode("utf-8"))
+            print(ll)
 
-        print()
+        print("")
 
     for i in renpy.config.lint_stats_callbacks:
         i()
 
-    print()
+    print("")
     if renpy.config.developer and (renpy.config.original_developer != "auto"):
         print("Remember to set config.developer to False before releasing.")
-        print()
+        print("")
 
     print("Lint is not a substitute for thorough testing. Remember to update Ren'Py")
     print("before releasing. New releases fix bugs and improve compatibility.")
 
+    if error_reported and args.error_code:
+        renpy.exports.quit(status=1)
+
     return False
