@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -22,24 +22,191 @@
 # This file contains functions used to help debug memory leaks. They aren't
 # called by default, but can be used when problems occur.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+
 import time
 import weakref
 import types
 import sys
 import collections
 import gc
+import inspect
 
 import renpy
 
 memory_log = renpy.log.open("memory")
 
+# Names that are intended to be constant containers and may be skipped during profiling.
+constant_containers = {
+    "renpy.game.script",
+    "renpy.display.image.images",
+    "renpy.display.image.image_attributes",
+    "renpy.python.py_compile_cache",
+    "renpy.python.old_py_compile_cache",
+    "renpy.python.store_dicts",
+    "renpy.python.store_modules",
+    "renpy.pyanalysis.ccache",
+    "renpy.gl2.live2d.common_cache",
+    "renpy.sl2.slast.scache",
+    "renpy.sl2.slast.ccache",
+    "renpy.sl2.slparser.all_statements",
+    "renpy.screenlang.all_statements",
+    "renpy.display.screen.screens_at_sort",
+    "renpy.display.screen.screens",
+    "renpy.display.screen.screens_by_name",
+}
+
+
+def print_garbage(gen):
+    """
+    Prints out the garbage after collecting a generation of memory.
+    """
+
+    print()
+    print("Garbage after collecting generation {}:".format(gen))
+
+    for i in gc.garbage:
+        prefix = ""
+        suffix = ""
+
+        if hasattr(i, "cell_contents"):
+            i = i.cell_contents
+            prefix = "cell: "
+
+        try:
+            suffix = " (" + inspect.getfile(i) + ")"
+        except:
+            pass
+
+        print(" -", prefix + repr(i)[:160] + suffix)
+
 
 def write(s):
     sys.stdout.write(s + "\n")
     memory_log.write("%s", s)
 
 
+def cycle_finder(o, name):
+    o_repr_cache = { }
+    paths = { }
+
+    edges = set()
+
+    def visit(old_ido, o, path):
+        ido = id(o)
+
+        if old_ido is not None:
+            edges.add((old_ido, ido, path))
+
+        if ido in o_repr_cache:
+            return
+
+        paths[ido] = path
+
+        if isinstance(o, (int, float, type(None), types.ModuleType, type)):
+            o_repr = repr(o)
+
+        elif isinstance(o, basestring):
+            if len(o) <= 80:
+                o_repr = repr(o).encode("utf-8")
+            else:
+                o_repr = repr(o[:80] + "...").encode("utf-8")
+
+        elif isinstance(o, (tuple, list)):
+            o_repr = "<" + o.__class__.__name__ + ">"
+
+        elif isinstance(o, dict):
+            o_repr = "<" + o.__class__.__name__ + ">"
+
+        elif isinstance(o, types.MethodType):
+            o_repr = "<method {0}.{1}>".format(o.__self__.__class__.__name__, o.__func__.__name__)
+
+        elif isinstance(o, object):
+            o_repr = "<{0}>".format(type(o).__name__)
+
+        else:
+            o_repr = "BAD TYPE <{0}>".format(type(o).__name__)
+
+        o_repr_cache[ido] = o_repr
+
+        if isinstance(o, (tuple, list)):
+            for i, oo in enumerate(o):
+                visit(ido, oo, "{0}[{1!r}]".format(path, i))
+
+        if isinstance(o, dict):
+            for k, v in o.items():
+                visit(ido, v, "{0}[{1!r}]".format(path, k))
+
+        elif isinstance(o, types.MethodType):
+            visit(ido, o.__self__, path + ".im_self")
+
+        else:
+
+            try:
+                reduction = o.__reduce_ex__(2)
+            except:
+                reduction = [ ]
+
+            # Gets an element from the reduction, or o if we don't have
+            # such an element.
+            def get(idx, default):
+                if idx < len(reduction) and reduction[idx] is not None:
+                    return reduction[idx]
+                else:
+                    return default
+
+            state = get(2, { })
+            if isinstance(state, dict):
+                for k, v in state.items():
+                    visit(ido, v, path + "." + k)
+            else:
+                visit(ido, state, path + ".__getstate__()")
+
+            for i, oo in enumerate(get(3, [])):
+                visit(ido, oo, "{0}[{1}]".format(path, i))
+
+            for i in get(4, []):
+
+                if len(i) != 2:
+                    continue
+
+                k, v = i
+
+                visit(ido, v, "{0}[{1!r}]".format(path, k))
+
+    visit(None, o, name)
+
+    while True:
+        left = set(i[0] for i in edges)
+        right = set(i[1] for i in edges)
+
+        leaves = right - left
+        roots = left - right
+
+        if (not leaves) and (not roots):
+            break
+
+        edges = set(i for i in edges if (i[1] not in leaves) if (i[0] not in roots))
+
+    while edges:
+        print()
+        print("Cycle:")
+
+        edge = list(edges)[0]
+
+        while edge in edges:
+            edges.remove(edge)
+            print("  ", edge)
+            print(" -", edge[2], "=", o_repr_cache[edge[1]])
+
+            relevant = [ i for i in edges if i[0] == edge[1] ]
+            if not relevant:
+                break
+
+            edge = relevant[0]
+
+
 def walk_memory(roots, seen=None):
     """
     Walks over memory, trying to account it to the objects in `roots`. Each
@@ -65,11 +232,6 @@ def walk_memory(roots, seen=None):
     # A map from root_name to total_size.
     size = collections.defaultdict(int)
 
-    def add(name, o):
-        """
-        Adds o to the worklist if it's not in seen.
-        """
-
     for name, o in roots:
         id_o = id(o)
 
@@ -84,7 +246,7 @@ def walk_memory(roots, seen=None):
     get_referents = gc.get_referents
     worklist_append = worklist.append
 
-    ignore_types = (types.ModuleType, types.ClassType, types.FunctionType)
+    ignore_types = (types.ModuleType, type, types.FunctionType)
 
     while worklist:
         name, o = worklist.pop(0)
@@ -106,7 +268,7 @@ def walk_memory(roots, seen=None):
     return size, seen
 
 
-def profile_memory_common(packages=[ "renpy", "store" ]):
+def profile_memory_common(packages=[ "renpy", "store" ], skip_constants=False):
     """
     Profiles object, surface, and texture memory used in the renpy and store
     packages.
@@ -136,12 +298,17 @@ def profile_memory_common(packages=[ "renpy", "store" ]):
             continue
 
         for name, o in mod.__dict__.items():
-            roots.append((mod_name + "." + name, o))
+            name = mod_name + "." + name
+
+            if skip_constants and name in constant_containers:
+                continue
+
+            roots.append((name, o))
 
     return walk_memory(roots)
 
 
-def profile_memory(fraction=1.0, minimum=0):
+def profile_memory(fraction=1.0, minimum=0, skip_constants=False):
     """
     :doc: memory
 
@@ -162,6 +329,10 @@ def profile_memory(fraction=1.0, minimum=0):
         If a name is accounted less than `minimum` bytes of memory, it will
         not be printed.
 
+    `skip_constants`
+        If True, the profiler will skip scanning of large Ren'Py's containers,
+        that are intended to be immutable after startup.
+
     As it has to scan all memory used by Ren'Py, this function may take a
     long time to complete.
     """
@@ -171,7 +342,7 @@ def profile_memory(fraction=1.0, minimum=0):
     write("Memory profile at " + time.ctime() + ":")
     write("")
 
-    usage = [ (v, k) for (k, v) in profile_memory_common()[0].items() ]
+    usage = [ (v, k) for (k, v) in profile_memory_common(skip_constants=skip_constants)[0].items() ]
     usage.sort()
 
     # The total number of bytes allocated.
@@ -192,11 +363,12 @@ def profile_memory(fraction=1.0, minimum=0):
     write("{:13,d} Total object, surface, and texture memory usage (in bytes).".format(total))
     write("")
 
+
 old_usage = { }
 old_total = 0
 
 
-def diff_memory(update=True):
+def diff_memory(update=True, skip_constants=False):
     """
     :doc: memory
 
@@ -209,6 +381,10 @@ def diff_memory(update=True):
     than one name, it's assigned to the name it's most directly reachable
     from.
 
+    `skip_constants`
+        If True, the profiler will skip scanning of large Ren'Py's containers,
+        that are intended to be immutable after startup.
+
     As it has to scan all memory used by Ren'Py, this function may take a
     long time to complete.
     """
@@ -221,12 +397,12 @@ def diff_memory(update=True):
     write("Memory diff at " + time.ctime() + ":")
     write("")
 
-    usage = profile_memory_common()[0]
+    usage = profile_memory_common(skip_constants=skip_constants)[0]
     total = sum(usage.values())
 
     diff = [ ]
 
-    for k, v in usage.iteritems():
+    for k, v in usage.items():
         diff.append((
             v - old_usage.get(k, 0),
             k))
@@ -278,8 +454,8 @@ def profile_rollback():
     # Walk the log, finding new roots and rollback information.
     for rb in log:
 
-        for store_name, store in rb.stores.iteritems():
-            for var_name, o in store.iteritems():
+        for store_name, store in rb.stores.items():
+            for var_name, o in store.items():
                 name = store_name + "." + var_name
                 id_o = id(o)
 
@@ -303,7 +479,7 @@ def profile_rollback():
 
     sizes = walk_memory(roots, seen)[0]
 
-    usage = [ (v, k) for (k, v) in sizes.iteritems() ]
+    usage = [ (v, k) for (k, v) in sizes.items() ]
     usage.sort()
 
     write("Total Bytes".rjust(13) + " " + "Per Rollback".rjust(13))
@@ -321,21 +497,17 @@ def profile_rollback():
     write("")
 
 
-################################################################################
-# Legacy memory debug functions
-################################################################################
-
 def find_parents(cls):
     """
     Finds the parents of every object of type `cls`.
     """
 
-    # GC to save memory.
-    gc.collect()
+    if gc.garbage:
+        del gc.garbage[:]
 
     objs = gc.get_objects()
 
-    def print_path(o):
+    def print_path(o, objs):
 
         prefix = ""
 
@@ -343,17 +515,20 @@ def find_parents(cls):
         queue = [ ]
         objects = [ ]
 
+        last = None
+
         for _i in range(30):
 
             objects.append(o)
+            last = o
 
-            print(prefix + str(id(o)), type(o), end=' ')
+            print(prefix + "%x" % id(o), "(%d referrers)" % len(gc.get_referrers(o)), type(o), end=' ')
 
             try:
                 if isinstance(o, dict) and "__name__" in o:
                     print("with name", o["__name__"])
                 else:
-                    print(repr(o))  # [:1000]
+                    print(repr(o))
             except:
                 print("Bad repr.")
 
@@ -397,13 +572,18 @@ def find_parents(cls):
 
             o, prefix = queue.pop()
 
+        for i in gc.get_referrers(last):
+            print(prefix + "<- %x" % id(i), type(i))
+
+        del objects[:]
+
     for o in objs:
         if isinstance(o, cls):
-            import random
-            if random.random() < .1:
 
-                print()
-                print("===================================================")
-                print()
+            print()
+            print("===================================================")
+            print()
 
-                print_path(o)
+            print_path(o, objs)
+
+    del objs[:]
