@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,6 +19,10 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+
+
 import renpy.display
 import contextlib
 
@@ -80,9 +84,6 @@ class Positional(object):
         if parser:
             parser.add(self)
 
-# Used to generate the documentation
-all_keyword_names = set()
-
 
 class Keyword(object):
     """
@@ -92,11 +93,10 @@ class Keyword(object):
     def __init__(self, name):
         self.name = name
 
-        all_keyword_names.add(self.name)
-
         if parser:
             parser.add(self)
 
+
 STYLE_PREFIXES = [
     '',
     'insensitive_',
@@ -119,9 +119,6 @@ class Style(object):
     def __init__(self, name):
         self.name = name
 
-        for j in STYLE_PREFIXES:
-            all_keyword_names.add(j + self.name)
-
         if parser:
             parser.add(self)
 
@@ -135,9 +132,6 @@ class PrefixStyle(object):
         self.prefix = prefix
         self.name = name
 
-        for j in STYLE_PREFIXES:
-            all_keyword_names.add(prefix + j + self.name)
-
         if parser:
             parser.add(self)
 
@@ -237,7 +231,7 @@ class Parser(object):
         and expr instances, and adjusts the line number.
         """
 
-        if isinstance(expr, unicode):
+        if isinstance(expr, str):
             expr = renpy.python.escape_unicode(expr)
 
         try:
@@ -260,7 +254,7 @@ class Parser(object):
         adjusts the line number. Returns a list of statements.
         """
 
-        if isinstance(code, unicode):
+        if isinstance(code, str):
             code = renpy.python.escape_unicode(code)
 
         try:
@@ -625,6 +619,7 @@ bar_properties = [ Style(i) for i in [
 box_properties = [ Style(i) for i in [
     "box_layout",
     "box_wrap",
+    "box_wrap_spacing",
     "box_reverse",
     "order_reverse",
     "spacing",
@@ -723,6 +718,7 @@ Keyword("default")
 Keyword("length")
 Keyword("allow")
 Keyword("exclude")
+Keyword("copypaste")
 Keyword("prefix")
 Keyword("suffix")
 Keyword("changed")
@@ -889,16 +885,19 @@ Keyword("droppable")
 Keyword("drag_raise")
 Keyword("dragged")
 Keyword("dropped")
+Keyword("drop_allowable")
 Keyword("drag_handle")
 Keyword("drag_joined")
 Keyword("clicked")
 Keyword("hovered")
 Keyword("unhovered")
+Keyword("mouse_drop")
 Style("child")
 add(ui_properties)
 add(position_properties)
 
 FunctionStatementParser("draggroup", "ui.draggroup", many)
+Keyword("min_overlap")
 add(ui_properties)
 add(position_properties)
 
@@ -921,6 +920,7 @@ class PassParser(Parser):
     def parse(self, l, name):
         return self.parse_exec("pass", l.number)
 
+
 PassParser("pass")
 
 
@@ -939,6 +939,7 @@ class DefaultParser(Parser):
 
         return self.parse_exec(code, l.number)
 
+
 DefaultParser("default")
 
 
@@ -980,6 +981,7 @@ class UseParser(Parser):
 
         return self.parse_exec(code, lineno)
 
+
 UseParser("use")
 
 
@@ -1043,6 +1045,7 @@ class IfParser(Parser):
 
         return [ rv ]
 
+
 IfParser("if")
 
 
@@ -1117,6 +1120,7 @@ class ForParser(Parser):
 
         return rv
 
+
 ForParser("for")
 
 
@@ -1142,6 +1146,7 @@ class PythonParser(Parser):
 
         return self.parse_exec(python_code, lineno)
 
+
 PythonParser("$", True)
 PythonParser("python", False)
 
@@ -1357,6 +1362,7 @@ class ScreenParser(Parser):
 
         return screen
 
+
 screen_parser = ScreenParser()
 screen_parser.add(all_statements)
 
