@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -22,7 +22,9 @@
 # This file contains code that is responsible for storing and executing a
 # Ren'Py script.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+
 import renpy
 
 import hashlib
@@ -34,7 +36,7 @@ import marshal
 import struct
 import zlib
 
-from cPickle import loads, dumps
+from renpy.compat.pickle import loads, dumps
 import shutil
 
 # The version of the dumped script.
@@ -47,7 +49,7 @@ BYTECODE_VERSION = 1
 MAGIC = imp.get_magic()
 
 # A string at the start of each rpycv2 file.
-RPYC2_HEADER = "RENPY RPC2"
+RPYC2_HEADER = b"RENPY RPC2"
 
 # A string
 BYTECODE_FILE = "cache/bytecode.rpyb"
@@ -107,13 +109,14 @@ class Script(object):
         renpy.game.script = self
 
         if os.path.exists(renpy.config.renpy_base + "/lock.txt"):
-            self.key = file(renpy.config.renpy_base + "/lock.txt", "rb").read()
+            self.key = open(renpy.config.renpy_base + "/lock.txt", "rb").read()
         else:
             self.key = None
 
         self.namemap = { }
         self.all_stmts = [ ]
         self.all_pycode = [ ]
+        self.all_pyexpr = [ ]
 
         # A list of statements that haven't been analyzed.
         self.need_analysis = [ ]
@@ -126,15 +129,15 @@ class Script(object):
         self.bytecode_dirty = False
 
         self.translator = renpy.translation.ScriptTranslator()
-
         self.init_bytecode()
+
         self.scan_script_files()
 
         self.translator.chain_translates()
 
         self.serial = 0
 
-        self.digest = hashlib.md5(renpy.version_only)
+        self.digest = hashlib.md5(renpy.version_only.encode("utf-8"))
 
         self.loaded_rpy = False
         self.backup_list = [ ]
@@ -151,7 +154,7 @@ class Script(object):
                 return None
 
         import __main__
-        backups = __main__.path_to_saves(renpy.config.gamedir, "backups")  # @UndefinedVariable
+        backups = __main__.path_to_saves(renpy.config.gamedir, "backups") # @UndefinedVariable
 
         if backups is None:
             return
@@ -190,12 +193,18 @@ class Script(object):
             if not os.path.exists(fn):
                 continue
 
-            short_fn = renpy.exports.fsencode(fn[len(renpy.config.gamedir)+1:])
+            short_fn = renpy.exports.fsencode(fn[len(renpy.config.gamedir) + 1:])
 
             base, ext = os.path.splitext(short_fn)
+
+            if PY2:
+                hex_checksum = checksum[:8].encode("hex")
+            else:
+                hex_checksum = checksum[:8].hex()
+
             target_fn = os.path.join(
                 backupdir,
-                base + "." + checksum[:8].encode("hex") + ext,
+                base + "." + hex_checksum + ext,
                 )
 
             if os.path.exists(target_fn):
@@ -226,7 +235,7 @@ class Script(object):
         # Similar, but for modules:
         self.module_files = [ ]
 
-        for dir, fn in dirlist:  # @ReservedAssignment
+        for dir, fn in dirlist: # @ReservedAssignment
 
             if fn.endswith(".rpy"):
                 if dir is None:
@@ -261,20 +270,31 @@ class Script(object):
 
         initcode = [ ]
 
-        for fn, dir in script_files:  # @ReservedAssignment
+        for fn, dir in script_files: # @ReservedAssignment
+            # Mitigate "busy script" warning from the browser
+            if renpy.emscripten:
+                import emscripten
+                emscripten.sleep(0)
+
+            # Pump the presplash window to prevent marking
+            # our process as unresponsive by OS
+            renpy.display.presplash.pump_window()
+
             self.load_appropriate_file(".rpyc", ".rpy", dir, fn, initcode)
 
         # Make the sort stable.
         initcode = [ (prio, index, code) for index, (prio, code) in
                      enumerate(initcode) ]
 
-        initcode.sort()
+        initcode.sort(key=lambda i: (i[0], i[1]))
 
         self.initcode = [ (prio, code) for prio, index, code in initcode ]
 
+        self.translator.chain_translates()
+
     def load_module(self, name):
 
-        files = [ (fn, dir) for fn, dir in self.module_files if fn == name ]  # @ReservedAssignment
+        files = [ (fn, dir) for fn, dir in self.module_files if fn == name ] # @ReservedAssignment
 
         if not files:
             raise Exception("Module %s could not be loaded." % name)
@@ -282,7 +302,7 @@ class Script(object):
         if len(files) > 2:
             raise Exception("Module %s ambiguous, multiple variants exist." % name)
 
-        fn, dir = files[0]  # @ReservedAssignment
+        fn, dir = files[0] # @ReservedAssignment
         initcode = [ ]
 
         self.load_appropriate_file(".rpymc", ".rpym", dir, fn, initcode)
@@ -306,7 +326,7 @@ class Script(object):
                 s.name = (fn, version, self.serial)
                 self.serial += 1
 
-    def merge_names(self, old_stmts, new_stmts):
+    def merge_names(self, old_stmts, new_stmts, used_names):
 
         old_stmts = collapse_stmts(old_stmts)
         new_stmts = collapse_stmts(new_stmts)
@@ -321,8 +341,9 @@ class Script(object):
                 old = old_stmts[oldl + i]
                 new = new_stmts[newl + i]
 
-                if new.name is None:
+                if (new.name is None) and (new.name not in used_names):
                     new.name = old.name
+                    used_names.add(new.name)
 
     def load_string(self, filename, filedata, linenumber=1):
         """
@@ -384,6 +405,10 @@ class Script(object):
         for i in stmts:
             i.get_children(all_stmts.append)
 
+        for i in all_stmts:
+            if isinstance(i, renpy.ast.RPY) and i.rest == ("python", "3"):
+                renpy.python.py3_files.add(i.filename)
+
         # Take the translations.
         self.translator.take_translates(all_stmts)
 
@@ -392,39 +417,52 @@ class Script(object):
             filename = renpy.parser.elide_filename(filename)
 
             if not all_stmts[0].filename.lower().endswith(filename.lower()):
-                filename += "c"
+
+                if filename[-1] != "c":
+                    filename += "c"
+
                 for i in all_stmts:
                     i.filename = filename
 
-        # Check for duplicate names.
-        if check_names and not renpy.mobile:
+        def check_name(node):
+
+            if not check_names:
+                return
+
+            if renpy.mobile:
+                return
+
             bad_name = None
             bad_node = None
             old_node = None
 
-            for node in all_stmts:
-                name = node.name
+            name = node.name
 
-                if name in self.namemap:
+            if name in self.namemap:
 
-                    bad_name = name
-                    bad_node = node
-                    old_node = self.namemap[name]
+                bad_name = name
+                bad_node = node
+                old_node = self.namemap[name]
 
-                    if not isinstance(bad_name, basestring):
+                if not isinstance(bad_name, basestring):
 
-                        raise ScriptError("Name %s is defined twice, at %s:%d and %s:%d." %
-                                          (repr(bad_name),
-                                           old_node.filename, old_node.linenumber,
-                                           bad_node.filename, bad_node.linenumber))
+                    raise ScriptError("Name %s is defined twice, at %s:%d and %s:%d." %
+                                      (repr(bad_name),
+                                       old_node.filename, old_node.linenumber,
+                                       bad_node.filename, bad_node.linenumber))
 
-                    else:
-                        self.duplicate_labels.append(
-                            u'The label {} is defined twice, at\n  File "{}", line {} and\n  File "{}", line {}.'.format(
-                                bad_name, old_node.filename, old_node.linenumber, bad_node.filename, bad_node.linenumber))
+                else:
 
-                # Add twice, so we can find duplicates in the same file.
-                self.namemap[name] = node
+                    if renpy.config.allow_duplicate_labels:
+                        return
+
+                    self.duplicate_labels.append(
+                        u'The label {} is defined twice, at File "{}", line {}:\n{}and File "{}", line {}:\n{}'.format(
+                            bad_name, old_node.filename, old_node.linenumber,
+                            renpy.parser.get_line_text(old_node.filename, old_node.linenumber),
+                            bad_node.filename, bad_node.linenumber,
+                            renpy.parser.get_line_text(old_node.filename, old_node.linenumber),
+                        ))
 
         self.update_bytecode()
 
@@ -432,6 +470,8 @@ class Script(object):
 
             name = node.name
 
+            check_name(node)
+
             # Add the name to the namemap.
             self.namemap[name] = node
 
@@ -470,7 +510,7 @@ class Script(object):
         f.seek(0, 2)
 
         start = f.tell()
-        data = zlib.compress(data, 9)
+        data = zlib.compress(data, 3)
         f.write(data)
 
         f.seek(len(RPYC2_HEADER) + 12 * (slot - 1), 0)
@@ -505,13 +545,13 @@ class Script(object):
             f.seek(0)
             data = f.read()
 
-            return data.decode("zlib")
+            return zlib.decompress(data)
 
         # RPYC2 path.
         pos = len(RPYC2_HEADER)
 
         while True:
-            header_slot, start, length = struct.unpack("III", header_data[pos:pos+12])
+            header_slot, start, length = struct.unpack("III", header_data[pos:pos + 12])
 
             if slot == header_slot:
                 break
@@ -536,16 +576,21 @@ class Script(object):
         # Generate translate nodes.
         renpy.translation.restructure(stmts)
 
-    def load_file(self, dir, fn):  # @ReservedAssignment
+    def load_file(self, dir, fn): # @ReservedAssignment
 
         if fn.endswith(".rpy") or fn.endswith(".rpym"):
 
             if not dir:
                 raise Exception("Cannot load rpy/rpym file %s from inside an archive." % fn)
 
+            base, _, game = dir.rpartition("/")
+            olddir = base + "/old-" + game
+
             fullfn = dir + "/" + fn
             rpycfn = fullfn + "c"
 
+            oldrpycfn = olddir + "/" + fn + "c"
+
             stmts = renpy.parser.parse(fullfn)
 
             data = { }
@@ -555,52 +600,51 @@ class Script(object):
             if stmts is None:
                 return data, [ ]
 
-            # See if we have a corresponding .rpyc file. If so, then
-            # we want to try to upgrade our .rpy file with it.
-            try:
-                self.record_pycode = False
+            used_names = set()
 
-                with open(rpycfn, "rb") as rpycf:
-                    bindata = self.read_rpyc_data(rpycf, 1)
-
-                old_data, old_stmts = loads(bindata)
-
-                self.merge_names(old_stmts, stmts)
-
-                del old_data
-                del old_stmts
-            except:
-                pass
-            finally:
-                self.record_pycode = True
-
-            self.assign_names(stmts, fullfn)
-
-            if not renpy.macapp:
+            for mergefn in [ oldrpycfn, rpycfn ]:
 
+                # See if we have a corresponding .rpyc file. If so, then
+                # we want to try to upgrade our .rpy file with it.
                 try:
-                    f = file(rpycfn, "wb")
+                    self.record_pycode = False
 
-                    self.write_rpyc_header(f)
-                    self.write_rpyc_data(f, 1, dumps((data, stmts), 2))
+                    with open(mergefn, "rb") as rpycf:
+                        bindata = self.read_rpyc_data(rpycf, 1)
+
+                    old_data, old_stmts = loads(bindata)
+
+                    self.merge_names(old_stmts, stmts, used_names)
+
+                    del old_data
+                    del old_stmts
                 except:
                     pass
+                finally:
+                    self.record_pycode = True
+
+            self.assign_names(stmts, renpy.parser.elide_filename(fullfn))
+
+            pickle_data_before_static_transforms = dumps((data, stmts), 2)
 
             self.static_transforms(stmts)
 
+            pickle_data_after_static_transforms = dumps((data, stmts), 2)
+
             if not renpy.macapp:
-
                 try:
-                    self.write_rpyc_data(f, 2, dumps((data, stmts), 2))
+                    with open(rpycfn, "wb") as f:
+                        self.write_rpyc_header(f)
+                        self.write_rpyc_data(f, 1, pickle_data_before_static_transforms)
+                        self.write_rpyc_data(f, 2, pickle_data_after_static_transforms)
 
-                    with open(fullfn, "rU") as fullf:
-                        rpydigest = hashlib.md5(fullf.read()).digest()
+                        with open(fullfn, "rU") as fullf:
+                            rpydigest = hashlib.md5(fullf.read()).digest()
 
-                    self.write_rpyc_md5(f, rpydigest)
-
-                    f.close()
+                        self.write_rpyc_md5(f, rpydigest)
                 except:
-                    pass
+                    import traceback
+                    traceback.print_exc()
 
             self.loaded_rpy = True
 
@@ -609,10 +653,7 @@ class Script(object):
             data = None
             stmts = None
 
-            f = renpy.loader.load(fn)
-
-            try:
-
+            with renpy.loader.load(fn) as f:
                 for slot in [ 2, 1 ]:
                     try:
                         bindata = self.read_rpyc_data(f, slot)
@@ -645,18 +686,16 @@ class Script(object):
                 if slot < 2:
                     self.static_transforms(stmts)
 
-            finally:
-                f.close()
-
         else:
             return None, None
 
         return data, stmts
 
-    def load_appropriate_file(self, compiled, source, dir, fn, initcode):  # @ReservedAssignment
+    def load_appropriate_file(self, compiled, source, dir, fn, initcode): # @ReservedAssignment
+        data = None
+
         # This can only be a .rpyc file, since we're loading it
         # from an archive.
-
         if dir is None:
 
             rpyfn = fn + source
@@ -666,10 +705,9 @@ class Script(object):
             if data is None:
                 raise Exception("Could not load from archive %s." % (lastfn,))
 
-            f = renpy.loader.load(fn + compiled)
-            f.seek(-hashlib.md5().digest_size, 2)
-            digest = f.read(hashlib.md5().digest_size)
-            f.close()
+            with renpy.loader.load(fn + compiled) as f:
+                f.seek(-hashlib.md5().digest_size, 2)
+                digest = f.read(hashlib.md5().digest_size)
 
         else:
 
@@ -681,8 +719,8 @@ class Script(object):
             renpy.loader.add_auto(rpyfn)
 
             if os.path.exists(rpyfn):
-                with open(rpyfn, "rU") as f:
-                    rpydigest = hashlib.md5(f.read()).digest()
+                with open(rpyfn, "rU", encoding="utf-8") as f:
+                    rpydigest = hashlib.md5(f.read().encode("utf-8")).digest()
             else:
                 rpydigest = None
 
@@ -696,10 +734,12 @@ class Script(object):
             except:
                 rpycdigest = None
 
+            digest = None
+
             if os.path.exists(rpyfn) and os.path.exists(rpycfn):
 
                 # Are we forcing a compile?
-                force_compile = renpy.game.args.compile  # @UndefinedVariable
+                force_compile = renpy.game.args.compile # @UndefinedVariable
 
                 # Use the source file here since it'll be loaded if it exists.
                 lastfn = rpyfn
@@ -716,6 +756,9 @@ class Script(object):
                             print("Could not load " + rpycfn)
 
                 except:
+                    renpy.display.log.write("While loading %r", rpycfn)
+                    renpy.display.log.exception()
+
                     if "RENPY_RPYC_EXCEPTIONS" in os.environ:
                         print("While loading", rpycfn)
                         raise
@@ -739,7 +782,8 @@ class Script(object):
 
                 digest = rpydigest
 
-            self.backup_list.append((rpyfn, digest))
+            if digest is not None:
+                self.backup_list.append((rpyfn, digest))
 
         if data is None:
             raise Exception("Could not load file %s." % lastfn)
@@ -748,9 +792,9 @@ class Script(object):
         if self.key is None:
             self.key = data['key']
         elif self.key != data['key']:
-            raise Exception( fn + " does not share a key with at least one .rpyc file. To fix, delete all .rpyc files, or rerun Ren'Py with the --lock option.")
+            raise Exception(fn + " does not share a key with at least one .rpyc file. To fix, delete all .rpyc files, or rerun Ren'Py with the --lock option.")
 
-        self.finish_load(stmts, initcode, filename=fn + source)
+        self.finish_load(stmts, initcode, filename=lastfn)
 
         self.digest.update(digest)
 
@@ -761,9 +805,11 @@ class Script(object):
 
         # Load the oldcache.
         try:
-            version, cache = loads(renpy.loader.load(BYTECODE_FILE).read().decode("zlib"))
-            if version == BYTECODE_VERSION:
-                self.bytecode_oldcache = cache
+            with renpy.loader.load(BYTECODE_FILE) as f:
+                version, cache = loads(zlib.decompress(f.read()))
+                if version == BYTECODE_VERSION:
+                    self.bytecode_oldcache = cache
+
         except:
             pass
 
@@ -773,12 +819,23 @@ class Script(object):
         cache. Clears out self.all_pycode.
         """
 
+        for i in self.all_pyexpr:
+            try:
+                renpy.python.py_compile(i, 'eval')
+            except:
+                pass
+
+        self.all_pyexpr = [ ]
+
         # Update all of the PyCode objects in the system with the loaded
         # bytecode.
         for i in self.all_pycode:
 
             key = i.get_hash() + MAGIC
 
+            if i.location[0] in renpy.python.py3_files:
+                key += b"_py3"
+
             code = self.bytecode_oldcache.get(key, None)
 
             if code is None:
@@ -792,6 +849,8 @@ class Script(object):
 
                     if i.mode == 'exec':
                         code = renpy.python.py_compile_exec_bytecode(i.source, filename=i.location[0], lineno=i.location[1])
+                    elif i.mode == 'hide':
+                        code = renpy.python.py_compile_hide_bytecode(i.source, filename=i.location[0], lineno=i.location[1])
                     elif i.mode == 'eval':
                         code = renpy.python.py_compile_eval_bytecode(i.source, filename=i.location[0], lineno=i.location[1])
 
@@ -802,10 +861,8 @@ class Script(object):
                     if text is None:
                         text = ''
 
-                    try:
-                        text = text.decode("utf-8")
-                    except:
-                        text = text.decode("latin-1")
+                    if not isinstance(text, str):
+                        text = text.decode("utf-8", "replace")
 
                     pem = renpy.parser.ParseError(
                         filename=e.filename,
@@ -835,7 +892,7 @@ class Script(object):
 
                 with open(fn, "wb") as f:
                     data = (BYTECODE_VERSION, self.bytecode_newcache)
-                    f.write(dumps(data, 2).encode("zlib"))
+                    f.write(zlib.compress(dumps(data, 2), 3))
             except:
                 pass
 
@@ -845,6 +902,9 @@ class Script(object):
         raises a ScriptError.
         """
 
+        if isinstance(label, renpy.parser.SubParse):
+            label = label.block[0].name
+
         label = renpy.config.label_overrides.get(label, label)
         original = label
 
@@ -864,10 +924,30 @@ class Script(object):
         Returns true if the label exists, or false otherwise.
         """
 
+        if isinstance(label, renpy.parser.SubParse):
+
+            if not label.block:
+                return False
+
+            label = label.block[0].name
+
         label = renpy.config.label_overrides.get(label, label)
 
         return label in self.namemap
 
+    def lookup_or_none(self, label):
+        """
+        Looks up the label if it exists, or returns None if it does not.
+        """
+
+        if label is None:
+            return None
+
+        if not self.has_label(label):
+            return None
+
+        return self.lookup(label)
+
     def analyze(self):
         """
         Analyzes all statements that need analysis.
