@@ -1,4 +1,4 @@
-# Copyright 2004-2015 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,6 +19,9 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+
 import pygame_sdl2 as pygame
 
 try:
@@ -30,11 +33,31 @@ import renpy.display
 import renpy.text.ftfont as ftfont
 import renpy.text.textsupport as textsupport
 
-ftfont.init() #@UndefinedVariable
+ftfont.init() # @UndefinedVariable
 
 WHITE = (255, 255, 255, 255)
 BLACK = (0, 0, 0, 255)
 
+
+def is_zerowidth(char):
+    if char == 0x200b: # Zero-width space.
+        return True
+
+    if char == 0x200c: # Zero-width non-joiner.
+        return True
+
+    if char == 0x200d: # Zero-width joiner.
+        return True
+
+    if char == 0x2060: # Word joiner.
+        return True
+
+    if char == 0xfeff: # Zero width non-breaking space.
+        return True
+
+    return False
+
+
 class ImageFont(object):
 
     # ImageFonts are expected to have the following fields defined by
@@ -52,7 +75,6 @@ class ImageFont(object):
     # offsets - The x and y offsets of each character.
     # chars - A map from a character to the surface containing that character.
 
-
     def glyphs(self, s):
 
         rv = [ ]
@@ -61,24 +83,30 @@ class ImageFont(object):
             return rv
 
         for c in s:
-            g = textsupport.Glyph() #@UndefinedVariable
+            g = textsupport.Glyph() # @UndefinedVariable
 
             g.character = ord(c)
-            g.ascent = self.height
+            g.ascent = self.baseline
             g.line_spacing = self.height
 
-            width = self.width.get(c, None)
-            if width is None:
-                raise Exception("Character {0!r} not found in image-based font.".format(c))
+            if not is_zerowidth(g.character):
 
-            g.width = self.width[c]
-            g.advance = self.advance[c]
+                width = self.width.get(c, None)
+                if width is None:
+                    raise Exception("Character {0!r} not found in image-based font.".format(c))
+
+                g.width = self.width[c]
+                g.advance = self.advance[c]
+
+            else:
+                g.width = 0
+                g.advance = 0
 
             rv.append(g)
 
         # Compute kerning.
         for i in range(len(s) - 1):
-            kern = self.kerns.get(s[i] + s[i+1], self.default_kern)
+            kern = self.kerns.get(s[i] + s[i + 1], self.default_kern)
             rv[i].advance += kern
 
         return rv
@@ -92,7 +120,11 @@ class ImageFont(object):
             return
 
         for g in glyphs:
-            c = unichr(g.character)
+
+            if not g.width:
+                continue
+
+            c = chr(g.character)
 
             cxo, cyo = self.offsets[c]
             x = g.x + xo + cxo
@@ -117,7 +149,8 @@ class SFont(ImageFont):
                  spacewidth,
                  default_kern,
                  kerns,
-                 charset):
+                 charset,
+                 baseline=None):
 
         super(SFont, self).__init__()
 
@@ -126,6 +159,7 @@ class SFont(ImageFont):
         self.default_kern = default_kern
         self.kerns = kerns
         self.charset = charset
+        self.baseline = baseline
 
     def load(self):
 
@@ -140,7 +174,11 @@ class SFont(ImageFont):
         sw, sh = surf.get_size()
         height = sh
         self.height = height # W0201
-        self.baseline = height # W0201
+        if self.baseline is None:
+            self.baseline = height # W0201
+        elif self.baseline < 0:
+            # Negative value is the distance from the bottom (vs top)
+            self.baseline = height + self.baseline # W0201
 
         # Create space characters.
         self.chars[u' '] = renpy.display.pgrender.surface((self.spacewidth, height), True)
@@ -223,7 +261,8 @@ class MudgeFont(ImageFont):
         surf = renpy.display.im.Image(self.filename).load(unscaled=True)
 
         # Parse the xml file.
-        tree = etree.fromstring(renpy.loader.load(self.xml).read())
+        with renpy.loader.load(self.xml) as f:
+            tree = etree.fromstring(f.read())
 
         height = 0
 
@@ -234,7 +273,7 @@ class MudgeFont(ImageFont):
             if char < 0:
                 continue
 
-            c = unichr(char)
+            c = chr(char)
             x = int(e.attrib["x"])
             y = int(e.attrib["y"])
             w = int(e.attrib["width"])
@@ -272,7 +311,6 @@ class MudgeFont(ImageFont):
         self.offsets[u'\u200b'] = (0, 0)
 
 
-
 def parse_bmfont_line(l):
     w = ""
     line = [ ]
@@ -298,9 +336,13 @@ def parse_bmfont_line(l):
     if w:
         line.append(w)
 
-    map = dict(i.split("=", 1) for i in line[1:]) #@ReservedAssignment
+    if not line:
+        line = [ "" ]
+
+    map = dict(i.split("=", 1) for i in line[1:]) # @ReservedAssignment
     return line[0], map
 
+
 class BMFont(ImageFont):
 
     def __init__(self, filename):
@@ -319,36 +361,38 @@ class BMFont(ImageFont):
 
         pages = { }
 
-        f = renpy.loader.load(self.filename)
-        for l in f:
+        with renpy.loader.load(self.filename) as f:
+            for l in f:
 
-            kind, args = parse_bmfont_line(l)
+                kind, args = parse_bmfont_line(l)
 
-            if kind == "common":
-                self.height = int(args["lineHeight"]) # W0201
-                self.baseline = int(args["base"]) # W0201
-            elif kind == "page":
-                pages[int(args["id"])] = renpy.display.im.Image(args["file"]).load(unscaled=True)
-            elif kind == "char":
-                c = unichr(int(args["id"]))
-                x = int(args["x"])
-                y = int(args["y"])
-                w = int(args["width"])
-                h = int(args["height"])
-                xo = int(args["xoffset"])
-                yo = int(args["yoffset"])
-                xadvance = int(args["xadvance"])
-                page = int(args["page"])
+                if kind == "common":
+                    self.height = int(args["lineHeight"]) # W0201
+                    self.baseline = int(args["base"]) # W0201
+                elif kind == "page":
+                    pages[int(args["id"])] = renpy.display.im.Image(args["file"]).load(unscaled=True)
+                elif kind == "char":
+                    c = chr(int(args["id"]))
+                    x = int(args["x"])
+                    y = int(args["y"])
+                    w = int(args["width"])
+                    h = int(args["height"])
+                    xo = int(args["xoffset"])
+                    yo = int(args["yoffset"])
+                    xadvance = int(args["xadvance"])
+                    page = int(args["page"])
 
-                ss = pages[page].subsurface((x, y, w, h))
-                ss = renpy.display.scale.surface_scale(ss)
+                    ss = pages[page].subsurface((x, y, w, h))
+                    ss = renpy.display.scale.surface_scale(ss)
 
-                self.chars[c] = ss
-                self.width[c] = w + xo
-                self.advance[c] = xadvance
-                self.offsets[c] = (xo, yo)
-
-        f.close()
+                    self.chars[c] = ss
+                    self.width[c] = w + xo
+                    self.advance[c] = xadvance
+                    self.offsets[c] = (xo, yo)
+                elif kind == "kerning":
+                    first = chr(int(args["first"]))
+                    second = chr(int(args["second"]))
+                    self.kerns[first + second] = int(args["amount"])
 
         if u'\u00a0' not in self.chars:
             self.chars[u'\u00a0'] = self.chars[u' ']
@@ -356,12 +400,12 @@ class BMFont(ImageFont):
             self.advance[u'\u00a0'] = self.advance[u' ']
             self.offsets[u'\u00a0'] = self.offsets[u' ']
 
-
         self.chars[u'\u200b'] = renpy.display.pgrender.surface((0, self.height), True)
         self.width[u'\u200b'] = 0
         self.advance[u'\u200b'] = 0
         self.offsets[u'\u200b'] = (0, 0)
 
+
 class ScaledImageFont(ImageFont):
     """
     Represents an imagefont scaled by a given factor.
@@ -376,14 +420,14 @@ class ScaledImageFont(ImageFont):
         self.baseline = scale(parent.baseline)
         self.default_kern = scale(parent.default_kern)
 
-        self.width = { k : scale(v) for k, v in parent.width.iteritems() }
-        self.advance = { k : scale(v) for k, v in parent.advance.iteritems() }
-        self.offsets = { k : (scale(v[0]), scale(v[1])) for k, v in parent.offsets.iteritems() }
-        self.kerns = { k : scale(v) for k, v in parent.kerns.iteritems() }
+        self.width = { k : scale(v) for k, v in parent.width.items() }
+        self.advance = { k : scale(v) for k, v in parent.advance.items() }
+        self.offsets = { k : (scale(v[0]), scale(v[1])) for k, v in parent.offsets.items() }
+        self.kerns = { k : scale(v) for k, v in parent.kerns.items() }
 
         self.chars = { }
 
-        for k, v in parent.chars.iteritems():
+        for k, v in parent.chars.items():
             w, h = v.get_size()
             nw = scale(w)
             nh = scale(h)
@@ -391,9 +435,8 @@ class ScaledImageFont(ImageFont):
 
 
 def register_sfont(name=None, size=None, bold=False, italics=False, underline=False,
-                   filename=None, spacewidth=10, default_kern=0, kerns={},
+                   filename=None, spacewidth=10, baseline=None, default_kern=0, kerns={},
                    charset=u"!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~"):
-
     """
     :doc: image_fonts
 
@@ -424,6 +467,13 @@ def register_sfont(name=None, size=None, bold=False, italics=False, underline=Fa
     `spacewidth`
         The width of a space character, an integer in pixels.
 
+    `baseline`
+        The distance from the top of the font to the baseline (the invisible
+        line letters sit on), an integer in pixels.  If this font is mixed with
+        other fonts, their baselines will be aligned.  Negative values indicate
+        distance from the bottom of the font instead, and ``None`` means the
+        baseline equals the height (i.e., is at the very bottom of the font).
+
     `default_kern`
         The default kern spacing between characters, in pixels.
 
@@ -431,7 +481,8 @@ def register_sfont(name=None, size=None, bold=False, italics=False, underline=Fa
         A map from two-character strings to the kern that should be used between
         those characters.
 
-    `charset` - The character set of the font. A string containing characters in
+    `charset`
+        The character set of the font. A string containing characters in
         the order in which they are found in the image. The default character
         set for a SFont is::
 
@@ -443,12 +494,12 @@ def register_sfont(name=None, size=None, bold=False, italics=False, underline=Fa
     if name is None or size is None or filename is None:
         raise Exception("When registering an SFont, the font name, font size, and filename are required.")
 
-    sf = SFont(filename, spacewidth, default_kern, kerns, charset)
+    sf = SFont(filename, spacewidth, default_kern, kerns, charset, baseline)
     image_fonts[(name, size, bold, italics)] = sf
 
-def register_mudgefont(name=None, size=None, bold=False, italics=False, underline=False,
-                   filename=None, xml=None, spacewidth=10, default_kern=0, kerns={}):
 
+def register_mudgefont(name=None, size=None, bold=False, italics=False, underline=False,
+                       filename=None, xml=None, spacewidth=10, default_kern=0, kerns={}):
     """
     :doc: image_fonts
 
@@ -500,9 +551,9 @@ def register_mudgefont(name=None, size=None, bold=False, italics=False, underlin
     mf = MudgeFont(filename, xml, spacewidth, default_kern, kerns)
     image_fonts[(name, size, bold, italics)] = mf
 
+
 def register_bmfont(name=None, size=None, bold=False, italics=False, underline=False,
                     filename=None):
-
     """
     :doc: image_fonts
 
@@ -542,9 +593,11 @@ def register_bmfont(name=None, size=None, bold=False, italics=False, underline=F
     bmf = BMFont(filename)
     image_fonts[(name, size, bold, italics)] = bmf
 
+
 # A map from face name to ftfont.FTFace
 face_cache = { }
 
+
 def load_face(fn):
 
     if fn in face_cache:
@@ -565,35 +618,37 @@ def load_face(fn):
         font_file = renpy.loader.load(fn)
     except IOError:
 
-        # Let's try to find the font on our own.
-        fonts = [ i.strip().lower() for i in fn.split(",") ]
+        if (not renpy.config.developer) or renpy.config.allow_sysfonts:
 
-        pygame.sysfont.initsysfonts()
+            # Let's try to find the font on our own.
+            fonts = [ i.strip().lower() for i in fn.split(",") ]
 
-        for v in pygame.sysfont.Sysfonts.itervalues():
-            if v is not None:
-                for _flags, ffn in v.iteritems():
-                    for i in fonts:
-                        if ffn.lower().endswith(i):
-                            font_file = file(ffn, "rb")
+            pygame.sysfont.initsysfonts()
+
+            for v in pygame.sysfont.Sysfonts.values():
+                if v is not None:
+                    for _flags, ffn in v.items():
+                        for i in fonts:
+                            if ffn.lower().endswith(i):
+                                font_file = open(ffn, "rb")
+                                break
+
+                        if font_file:
                             break
 
-                    if font_file:
-                        break
-
-            if font_file:
-                break
-
+                if font_file:
+                    break
 
     if font_file is None:
         raise Exception("Could not find font {0!r}.".format(orig_fn))
 
-    rv = ftfont.FTFace(font_file, index) #@UndefinedVariable
+    rv = ftfont.FTFace(font_file, index, orig_fn) # @UndefinedVariable
 
     face_cache[orig_fn] = rv
 
     return rv
 
+
 # Caches of fonts.
 image_fonts = { }
 
@@ -606,6 +661,7 @@ font_cache = { }
 # The last_scale we last accessed fonts at. (Used to clear caches.)
 last_scale = 1.0
 
+
 def get_font(fn, size, bold, italics, outline, antialias, vertical, hinting, scale):
 
     # If the scale changed, invalidate caches of scaled fonts.
@@ -644,12 +700,13 @@ def get_font(fn, size, bold, italics, outline, antialias, vertical, hinting, sca
 
     # Load a TTF.
     face = load_face(fn)
-    rv = ftfont.FTFont(face, int(size * scale), bold, italics, outline, antialias, vertical, hinting) #@UndefinedVariable
+    rv = ftfont.FTFont(face, int(size * scale), bold, italics, outline, antialias, vertical, hinting) # @UndefinedVariable
 
     font_cache[key] = rv
 
     return rv
 
+
 def free_memory():
     """
     Clears the font cache.
@@ -657,13 +714,15 @@ def free_memory():
 
     scaled_image_fonts.clear()
     font_cache.clear()
-    face_cache.clear()
 
 
-def load_image_fonts():
-    for i in image_fonts.itervalues():
+def load_fonts():
+    for i in image_fonts.values():
         i.load()
 
+    for i in renpy.config.preload_fonts:
+        load_face(i)
+
 
 class FontGroup(object):
     """
@@ -673,18 +732,19 @@ class FontGroup(object):
     A group of fonts that can be used as a single font.
     """
 
+    # For compatibility with older instances.
+    char_map = dict()
+
     def __init__(self):
 
-        # A list of font names we know of.
-        self.fonts = [ ]
+        # A map from character index to font name. None is used for
+        # the default font.
+        self.map = { }
 
-        # A map from character to the index of the font it's part of.
-        self.cache = { }
+        # A map from character number to character number, used to implement remap.
+        self.char_map = { }
 
-        # A list of (index, start, end) tuples.
-        self.patterns = [ ]
-
-    def add(self, font, start, end):
+    def add(self, font, start, end, target=None, target_increment=False):
         """
         :doc: font_group
 
@@ -692,11 +752,25 @@ class FontGroup(object):
 
         `start`
             The start of the range. This may be a single-character string, or
-            an integer giving a unicode code point.
+            an integer giving a unicode code point. If start is None, then the
+            font is used as the default.
 
         `end`
             The end of the range. This may be a single-character string, or an
-            integer giving a unicode code point.
+            integer giving a unicode code point. This is ignored if start is
+            None.
+
+        `target`
+            If given, associates the given range of characters with specific
+            characters from the given font, depending on target_increment.
+            This may be a single-character string, or an integer giving a
+            unicode code point. This is ignored if the character had already
+            been added.
+
+        `target_increment`
+            If True, the [start, end] range is mapped to the
+            [target, target+end-start] range. If False, every character from the
+            range is associated with the target character.
 
         When multiple .add() calls include the same character, the first call
         takes precedence.
@@ -705,21 +779,78 @@ class FontGroup(object):
         chained together.
         """
 
+        if start is None:
+
+            if isinstance(font, FontGroup):
+                for k, v in font.map.items():
+                    if k not in self.map:
+                        self.map[k] = v
+            else:
+                if None not in self.map:
+                    self.map[None] = font
+
+            return self
+
         if not isinstance(start, int):
             start = ord(start)
 
         if not isinstance(end, int):
             end = ord(end)
 
+        if target and not isinstance(target, int):
+            target = ord(target)
+
         if end < start:
             raise Exception("In FontGroup.add, the start of a character range must be before the end of the range.")
 
-        if font not in self.fonts:
-            self.fonts.append(font)
+        for i in range(start, end + 1):
+            if i not in self.map:
+                self.map[i] = font
 
-        index = self.fonts.index(font)
+                if target is not None:
+                    self.char_map[i] = target
 
-        self.patterns.append((index, start, end))
+                    if target_increment:
+                        target += 1
+
+        return self
+
+    def remap(self, cha, target):
+        """
+        :doc: font_group
+
+        Remaps one or a set of characters to a single target character.
+
+        `cha`
+            The character or characters to remap. This may be a single-character
+            string, or an integer giving a unicode code point, or an iterable of
+            either.
+
+        `target`
+            The character to remap to. This may be a single-character string, or
+            an integer giving a unicode code point.
+
+        Any given character having already been remapped (either with add or with
+        remap) will be ignored. However, if the FontGroup has no default font, any
+        given character must have been previously added.
+
+        This method also returns the FontGroup, for the same reasons.
+        """
+
+        if isinstance(cha, (int)) or isinstance(cha, (str, bytes)) and len(cha) == 1:
+            cha = (cha,)
+
+        if not isinstance(target, int):
+            target = ord(target)
+
+        for i in cha:
+            if not isinstance(i, int):
+                i = ord(i)
+            if not ((None in self.map) or (i in self.map)):
+                raise Exception("Character U+{0:04x} has no font in this FontGroup".format(i))
+            if not i in self.char_map:
+                # means the character has not already been remapped
+                self.char_map[i] = target
 
         return self
 
@@ -731,33 +862,31 @@ class FontGroup(object):
         mark = 0
         pos = 0
 
-        old_index = 0
+        old_font = None
 
-        cache = self.cache
+        if self.char_map:
+            s = [ ord(i) for i in s ]
+            s = "".join(chr(self.char_map.get(i, i)) for i in s)
 
-        for c in s:
+        for i, c in enumerate(s):
 
-            index = cache.get(c, None)
+            n = ord(c)
 
-            if index is None:
-                n = ord(c)
+            font = self.map.get(ord(c), None)
 
-                for index, start, end in self.patterns:
-                    if start <= n <= end:
-                        break
-                else:
+            if font is None:
+                font = self.map.get(None, None)
+
+                if font is None:
                     raise Exception("Character U+{0:04x} not found in FontGroup".format(n))
 
-                cache[c] = index
-
-            if index != old_index:
+            if font != old_font:
                 if pos:
-                    yield self.fonts[old_index], s[mark:pos]
+                    yield old_font, s[mark:pos]
 
-                old_index = index
+                old_font = font
                 mark = pos
 
             pos += 1
 
-        yield self.fonts[old_index], s[mark:]
-
+        yield font, s[mark:]
