@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,22 +19,22 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-from __future__ import print_function, unicode_literals
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
 
 import os
 import re
-import codecs
 
 import renpy.translation
 
 ################################################################################
 
 STRING_RE = r"""(?x)
-\b__?\s*\(\s*[uU]?(
-\"\"\"(?:\\.|\"{1,2}|[^\\"])*?\"\"\"
-|'''(?:\\.|\'{1,2}|[^\\'])*?'''
-|"(?:\\.|[^\\"])*"
-|'(?:\\.|[^\\'])*'
+\b_[_p]?\s*\(\s*[uU]?(
+\"\"\"(?:\\.|\\\n|\"{1,2}|[^\\"])*?\"\"\"
+|'''(?:\\.|\\\n|\'{1,2}|[^\\'])*?'''
+|"(?:\\.|\\\n|[^\\"])*"
+|'(?:\\.|\\\n|[^\\'])*'
 )\s*\)
 """
 
@@ -46,7 +46,6 @@ REGULAR_PRIORITIES = [
     ("", 100, "launcher.rpy"),
 ]
 
-
 COMMON_PRIORITIES = [
     ("_compat/", 420, "obsolete.rpy"),
     ("_layout/", 410, "obsolete.rpy"),
@@ -119,7 +118,7 @@ def scan_strings(filename):
 
     rv = [ ]
 
-    for line, s in renpy.game.script.translator.additional_strings[filename]:  # @UndefinedVariable
+    for line, s in renpy.game.script.translator.additional_strings[filename]: # @UndefinedVariable
         rv.append(String(filename, line, s, False))
 
     for _filename, lineno, text in renpy.parser.list_logical_lines(filename):
@@ -127,11 +126,16 @@ def scan_strings(filename):
         for m in re.finditer(STRING_RE, text):
 
             s = m.group(1)
+            s = s.replace('\\\n', "")
+
             if s is not None:
                 s = s.strip()
                 s = "u" + s
                 s = eval(s)
 
+                if m.group(0).startswith("_p"):
+                    s = renpy.minstore._p(s)
+
                 if s:
                     rv.append(String(filename, lineno, s, False))
 
@@ -148,7 +152,7 @@ def scan_comments(filename):
     comment = [ ]
     start = 0
 
-    with codecs.open(filename, "r", "utf-8") as f:
+    with open(filename, "r", encoding="utf-8") as f:
         lines = [ i.rstrip() for i in f.read().replace(u"\ufeff", "").split('\n') ]
 
     for i, l in enumerate(lines):
