@@ -689,14 +689,6 @@ class ChoiceActionBase(Action):
         else:
             self.block_all = block_all
 
-        self.chosen = None
-
-        if self.location:
-            self.chosen = renpy.game.persistent._chosen # @UndefinedVariable
-
-            if self.chosen is None:
-                self.chosen = renpy.game.persistent._chosen = { }
-
         # The arguments passed to a menu choice.
         self.args = args
         self.kwargs = kwargs
@@ -709,6 +701,13 @@ class ChoiceActionBase(Action):
         roll_forward = renpy.exports.roll_forward_info()
         return renpy.exports.in_fixed_rollback() and roll_forward == self.value
 
+    @property
+    def chosen(self):
+        if not self.location:
+            return None
+
+        return renpy.game.persistent._chosen
+
     def get_chosen(self):
         if self.chosen is None:
             return False
@@ -991,6 +990,7 @@ def _textbutton(label, clicked=None, style=None, text_style=None, substitute=Tru
     rv = renpy.display.behavior.Button(style=style, clicked=clicked, **button_kwargs)
     text = renpy.text.text.Text(label, style=text_style, substitute=substitute, scope=scope, **text_kwargs)
     rv.add(text)
+    rv._main = text
     rv._composite_parts = [ text ]
     return rv
 
